"""_summary_."""
class A:
    def __init__(self):
        """_summary_."""
        self._x = None
        self.test1 = "test"
        self.test2 = None
        self.test2 = 1
        self.test1 = "a"
        self.test3 = self.test4 = None

    @property
    def x(self) -> str | None:
        """I'm the 'x' property.

        Returns
        -------
        str | None
            _description_
        """
        return self._x

    @x.setter
    def x(self, value):
        """_summary_.

        Parameters
        ----------
        value : _type_
            _description_
        """
        self._x = value

    @staticmethod
    def a(self, a):
        """_summary_.

        Parameters
        ----------
        a : _type_
            _description_
        """
        pass

    @classmethod
    def b(self, b):
        """_summary_.

        Parameters
        ----------
        b : _type_
            _description_
        """
        pass

    def c(self, c):
        """_summary_.

        Parameters
        ----------
        c : _type_
            _description_
        """
        pass

    def d(self, pos, /, a: "annotation", b: int, c: int, *args: list, d: int = 5, e="test", **kwargs: dict):
        """_summary_.

        Parameters
        ----------
        pos : _type_
            _description_
        a : 'annotation'
            _description_
        b : int
            _description_
        c : int
            _description_
        *args : list
            _description_
        d : int
            _description_ (Default value = 5)
        e : _type_
            _description_ (Default value = 'test')
        **kwargs : dict
            _description_
        """
        pass

class B:
    """_summary_."""
    def __init__(self):
        """_summary_."""
        self._x = None
        self.test1 = "test"
        self.test2 = None
        self.test2 = 1
        self.test1 = "a"
        self.test3 = self.test4 = None

    @property
    def x(self) -> str | None:
        """I'm the 'x' property.

        Returns
        -------
        str | None
            _description_
        """
        return self._x

    @x.setter
    def x(self, value):
        """_summary_.

        Parameters
        ----------
        value : _type_
            _description_
        """
        self._x = value

    @staticmethod
    def a(self, a):
        """_summary_.

        Parameters
        ----------
        a : _type_
            _description_
        """
        pass

    @classmethod
    def b(self, b):
        """_summary_.

        Parameters
        ----------
        b : _type_
            _description_
        """
        pass

    def c(self, c):
        """_summary_.

        Parameters
        ----------
        c : _type_
            _description_
        """
        pass

class C:
    """_summary_.

    Attributes
    ----------
    a
        some description
    b : int
    """
    def __init__(self):
        """_summary_."""
        self.a = None
        self.b = None
        self.c = None


"""Test for utility functions."""

from pymend.docstring_parser.common import DocstringReturns
from pymend.docstring_parser.util import combine_docstrings


def test_combine_docstrings() -> None:
    """Test combine_docstrings wrapper."""

    def fun1(arg_a, arg_b, arg_c, arg_d) -> None:  # noqa: ANN001
        """short_description: fun1.

        :param arg_a: fun1
        :param arg_b: fun1
        :return: fun1
        """
        assert arg_a
        assert arg_b
        assert arg_c
        assert arg_d

    def fun2(arg_b, arg_c, arg_d, arg_e) -> None:  # noqa: ANN001
        """short_description: fun2.

        long_description: fun2

        :param arg_b: fun2
        :param arg_c: fun2
        :param arg_e: fun2
        """
        assert arg_b
        assert arg_c
        assert arg_d
        assert arg_e

    @combine_docstrings(fun1, fun2)
    def decorated1(arg_a, arg_b, arg_c, arg_d, arg_e, arg_f) -> None:  # noqa: ANN001
        """
        :param arg_e: decorated
        :param arg_f: decorated
        """  # noqa: D205
        assert arg_a
        assert arg_b
        assert arg_c
        assert arg_d
        assert arg_e
        assert arg_f

    assert decorated1.__doc__ == (
        "short_description: fun2.\n"
        "\n"
        "long_description: fun2\n"
        "\n"
        ":param arg_a: fun1\n"
        ":param arg_b: fun1\n"
        ":param arg_c: fun2\n"
        ":param arg_e: fun2\n"
        ":param arg_f: decorated\n"
        ":returns: fun1"
    )

    @combine_docstrings(fun1, fun2, exclude=[DocstringReturns])
    def decorated2(arg_a, arg_b, arg_c, arg_d, arg_e, arg_f) -> None:  # noqa: ANN001
        assert arg_a
        assert arg_b
        assert arg_c
        assert arg_d
        assert arg_e
        assert arg_f

    assert decorated2.__doc__ == (
        "short_description: fun2.\n"
        "\n"
        "long_description: fun2\n"
        "\n"
        ":param arg_a: fun1\n"
        ":param arg_b: fun1\n"
        ":param arg_c: fun2\n"
        ":param arg_e: fun2"
    )



"""Module for parsing input file and walking ast."""

import ast
import re
from typing import Optional, Union

from typing_extensions import TypeGuard

from .docstring_parser.attrdoc import ast_unparse
from .types import (
    ClassDocstring,
    DocstringInfo,
    ElementDocstring,
    FixerSettings,
    FunctionBody,
    FunctionDocstring,
    FunctionSignature,
    ModuleDocstring,
    NodeOfInterest,
    Parameter,
    ReturnValue,
)

__author__ = "J-E. Nitschke"
__copyright__ = "Copyright 2023-2023"
__licence__ = "GPL3"
__version__ = "1.0.0"
__maintainer__ = "J-E. Nitschke"


class AstAnalyzer:
    """Walk ast and extract module, class and function information."""

    def __init__(self, file_content: str, *, settings: FixerSettings) -> None:
        """Initialize the Analyzer with the file contents.

        The only reason this is a class is to have the raw
        file_contents available at any point of the analysis to double check
        something. Currently used for the module docstring and docstring
        modifiers.

        Parameters
        ----------
        file_content : str
            File contents to store.
        settings : FixerSettings
            Settings for what to fix and when.
        """
        self.file_content = file_content
        self.settings = settings

    def parse_from_ast(
        self,
    ) -> list[ElementDocstring]:
        """Walk AST of the input file extract info about module, classes and functions.

        For the module and classes and the raw docstring
        and its line numbers are extracted.

        For functions the raw docstring and its line numbers are extracted.
        Additionally the signature is parsed for parameters and return value.

        Returns
        -------
        list[ElementDocstring]
            List of information about module, classes and functions.

        Raises
        ------
        AssertionError
            If the source file content could not be parsed into an ast.
        """
        nodes_of_interest: list[ElementDocstring] = []
        try:
            file_ast = ast.parse(self.file_content)
        except Exception as exc:  # noqa: BLE001
            msg = f"Failed to parse source file AST: {exc}\n"
            raise AssertionError(msg) from exc
        for node in ast.walk(file_ast):
            if isinstance(node, ast.Module):
                nodes_of_interest.append(self.handle_module(node))
            elif isinstance(node, ast.ClassDef):
                if node.name in self.settings.ignored_classes:
                    continue
                nodes_of_interest.append(self.handle_class(node))
            elif isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
                if (
                    any(
                        name.id in self.settings.ignored_decorators
                        for name in node.decorator_list
                        if isinstance(name, ast.Name)
                    )
                    or node.name in self.settings.ignored_functions
                ):
                    continue
                nodes_of_interest.append(self.handle_function(node))
        return nodes_of_interest

    def handle_module(self, module: ast.Module) -> ModuleDocstring:
        """Extract information about module.

        Parameters
        ----------
        module : ast.Module
            Node representing the full module.

        Returns
        -------
        ModuleDocstring
            Docstring representation for the module.
        """
        docstring_info = self.get_docstring_info(module)
        if docstring_info is None:
            docstring_line = self._get_docstring_line()
            return ModuleDocstring(
                "Module",
                docstring="",
                lines=(docstring_line, docstring_line),
                modifier="",
                issues=[],
            )
        return ModuleDocstring(
            name=docstring_info.name,
            docstring=docstring_info.docstring,
            lines=docstring_info.lines,
            modifier=docstring_info.modifier,
            issues=docstring_info.issues,
        )

    def handle_class(self, cls: ast.ClassDef) -> ClassDocstring:
        """Extract information about class docstring.

        Parameters
        ----------
        cls : ast.ClassDef
            Node representing a class definition.

        Returns
        -------
        ClassDocstring
            Docstring representation for a class.
        """
        docstring = self.handle_elem_docstring(cls)
        attributes, methods = self.handle_class_body(cls)
        return ClassDocstring(
            name=docstring.name,
            docstring=docstring.docstring,
            lines=docstring.lines,
            modifier=docstring.modifier,
            issues=docstring.issues,
            attributes=attributes,
            methods=methods,
        )

    def handle_function(
        self,
        func: Union[ast.FunctionDef, ast.AsyncFunctionDef],
    ) -> FunctionDocstring:
        """Extract information from signature and docstring.

        Parameters
        ----------
        func : Union[ast.FunctionDef, ast.AsyncFunctionDef]
            Node representing a function definition.

        Returns
        -------
        FunctionDocstring
            Docstring representation of a function.
        """
        docstring = self.handle_elem_docstring(func)
        signature = self.handle_function_signature(func)
        body = self.handle_function_body(func)
        length = len(func.body) if func.body else 0
        if length and ast.get_docstring(func):
            length -= 1
        return FunctionDocstring(
            name=docstring.name,
            docstring=docstring.docstring,
            lines=docstring.lines,
            modifier=docstring.modifier,
            issues=docstring.issues,
            signature=signature,
            body=body,
            length=length,
        )

    def handle_elem_docstring(
        self,
        elem: Union[ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef],
    ) -> DocstringInfo:
        """Extract information about the docstring of the function.

        Parameters
        ----------
        elem : Union[ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef]
            Element representing a function or class definition.

        Returns
        -------
        DocstringInfo
            Return general information about the docstring of the element.

        Raises
        ------
        ValueError
            If the element did not have a body at all. This should not happen
            for valid functions or classes.
        """
        docstring_info = self.get_docstring_info(elem)
        if docstring_info is None:
            if not elem.body:
                msg = "Function body was unexpectedly completely empty."
                raise ValueError(msg)
            lines = (elem.body[0].lineno, elem.body[0].lineno)
            return DocstringInfo(
                name=elem.name, docstring="", lines=lines, modifier="", issues=[]
            )
        return docstring_info

    def get_docstring_info(self, node: NodeOfInterest) -> Optional[DocstringInfo]:
        """Get docstring and line number if available.

        Parameters
        ----------
        node : NodeOfInterest
            Get general information about the docstring of any node
            if interest.

        Returns
        -------
        Optional[DocstringInfo]
            Information about the docstring if the element contains one.
            Or `None` if there was no docstring at all.

        Raises
        ------
        ValueError
            If the first element of the body is not a docstring after
            `ast.get_docstring()` returned one.
        """
        if ast.get_docstring(node):
            if not (
                node.body
                and isinstance(first_element := node.body[0], ast.Expr)
                and isinstance(docnode := first_element.value, ast.Constant)
                and isinstance(docnode.value, str)
            ):
                msg = (
                    "Expected first entry in body to be the "
                    "docstring, but found nothing or something else."
                )
                raise ValueError(msg)
            modifier = self._get_modifier(
                self.file_content.splitlines()[docnode.lineno - 1]
            )
            return DocstringInfo(
                # Can not use DefinitionNodes in isinstance checks before 3.10
                name=node.name
                if isinstance(
                    node, (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef)
                )
                else "Module",
                docstring=str(docnode.value),
                lines=(docnode.lineno, docnode.end_lineno),
                modifier=modifier,
                issues=[],
            )
        return None

    def _get_modifier(self, line: str) -> str:
        """Get the string modifier from the start of a docstring.

        Parameters
        ----------
        line : str
            Line to check

        Returns
        -------
        str
            Modifier(s) of the string.
        """
        line = line.strip()
        delimiters = ['"""', "'''"]
        modifiers = ["r", "u", "f"]
        if not line:
            return ""
        if line[:3] in delimiters:
            return ""
        if line[0] in modifiers and line[1:4] in delimiters:
            return line[0]
        if line[0] in modifiers and line[1] in modifiers and line[2:5] in delimiters:
            return line[:2]
        return ""

    def _get_docstring_line(self) -> int:
        """Get the line where the module docstring should start.

        Returns
        -------
        int
            Starting line (starts at 1) of the docstring.
        """
        shebang_encoding_lines = 2
        for index, line in enumerate(
            self.file_content.splitlines()[:shebang_encoding_lines]
        ):
            if not self.is_shebang_or_pragma(line):
                # List indices start at 0 but file lines are counted from 1
                return index + 1
        return shebang_encoding_lines + 1

    def handle_class_body(self, cls: ast.ClassDef) -> tuple[list[Parameter], list[str]]:
        """Extract attributes and methods from class body.

        Will walk the AST of the ClassDef node and add each function encountered
        as a method.

        If the `__init__` method is encountered walk its body for attribute
        definitions.

        Parameters
        ----------
        cls : ast.ClassDef
            Node representing a class definition.

        Returns
        -------
        attributes : list[Parameter]
            List of the parameters that make up the classes attributes.
        methods : list[str]
            List of the method names in the class.
        """
        attributes: list[Parameter] = []
        methods: list[str] = []
        for node in cls.body:
            if not isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
                continue
            # Extract attributes from init method.
            if node.name == "__init__":
                attributes.extend(self._get_attributes_from_init(node))
            # Skip dunder methods for method extraction
            if node.name.startswith("__") and node.name.endswith("__"):
                continue
            # Optionally skip private methods.
            if self.settings.ignore_privates and node.name.startswith("_"):
                continue
            # Handle properties as attributes
            if "property" in {
                decorator.id
                for decorator in node.decorator_list
                if isinstance(decorator, ast.Name)
            }:
                return_value = self.get_return_value_sig(node)
                attributes.append(Parameter(node.name, return_value.type_name, None))
            # Handle normal methods except for those with some specific decorators
            # Like statismethod, classmethod, property or getters/setters.
            elif not self._has_excluding_decorator(node):
                methods.append(self._get_method_signature(node))
            # Exclude some like staticmethods and properties

        # Remove duplicates from attributes while maintaining order
        return list(Parameter.uniquefy(attributes)), methods

    def handle_function_signature(
        self,
        func: Union[ast.FunctionDef, ast.AsyncFunctionDef],
    ) -> FunctionSignature:
        """Extract information about the signature of the function.

        Parameters
        ----------
        func : Union[ast.FunctionDef, ast.AsyncFunctionDef]
            Node representing a function definition

        Returns
        -------
        FunctionSignature
            Information extracted from the function signature
        """
        parameters = self.get_parameters_sig(func)
        if parameters and parameters[0].arg_name == "self":
            parameters.pop(0)
        return_value = self.get_return_value_sig(func)
        return FunctionSignature(parameters, return_value)

    def handle_function_body(
        self,
        func: Union[ast.FunctionDef, ast.AsyncFunctionDef],
    ) -> FunctionBody:
        """Check the function body for yields, raises and value returns.

        Parameters
        ----------
        func : Union[ast.FunctionDef, ast.AsyncFunctionDef]
            Node representing a function definition

        Returns
        -------
        FunctionBody
            Information extracted from the function body.
        """
        returns: set[tuple[str, ...]] = set()
        returns_value = False
        yields: set[tuple[str, ...]] = set()
        yields_value = False
        raises: list[str] = []
        for node in ast.walk(func):
            if isinstance(node, ast.Return) and node.value is not None:
                returns_value = True
                if isinstance(node.value, ast.Tuple) and all(
                    isinstance(value, ast.Name) for value in node.value.elts
                ):
                    returns.add(self._get_ids_from_returns(node.value.elts))
            elif isinstance(node, (ast.Yield, ast.YieldFrom)):
                yields_value = True
                if (
                    isinstance(node, ast.Yield)
                    and isinstance(node.value, ast.Tuple)
                    and all(isinstance(value, ast.Name) for value in node.value.elts)
                ):
                    yields.add(self._get_ids_from_returns(node.value.elts))
            elif isinstance(node, ast.Raise):
                pascal_case_regex = r"^(?:[A-Z][a-z]+)+$"
                if not node.exc:
                    raises.append("")
                elif isinstance(node.exc, ast.Name) and re.match(
                    pascal_case_regex, node.exc.id
                ):
                    raises.append(node.exc.id)
                elif (
                    isinstance(node.exc, ast.Call)
                    and isinstance(node.exc.func, ast.Name)
                    and re.match(pascal_case_regex, node.exc.func.id)
                ):
                    raises.append(node.exc.func.id)
                else:
                    raises.append("")
        return FunctionBody(
            returns_value=returns_value,
            returns=returns,
            yields_value=yields_value,
            yields=yields,
            raises=raises,
        )

    def get_return_value_sig(
        self, func: Union[ast.FunctionDef, ast.AsyncFunctionDef]
    ) -> ReturnValue:
        """Get information about return value from signature.

        Parameters
        ----------
        func : Union[ast.FunctionDef, ast.AsyncFunctionDef]
            Node representing a function definition

        Returns
        -------
        ReturnValue
            Return information extracted from the function signature.
        """
        return ReturnValue(type_name=ast_unparse(func.returns))

    def get_parameters_sig(
        self, func: Union[ast.FunctionDef, ast.AsyncFunctionDef]
    ) -> list[Parameter]:
        """Get information about function parameters.

        Parameters
        ----------
        func : Union[ast.FunctionDef, ast.AsyncFunctionDef]
            Node representing a function definition

        Returns
        -------
        list[Parameter]
            Parameter information from the function signature.
        """
        arguments: list[Parameter] = []
        pos_defaults = self.get_padded_args_defaults(func)

        pos_only_args = [
            Parameter(arg.arg, ast_unparse(arg.annotation), None)
            for arg in func.args.posonlyargs
        ]
        arguments += pos_only_args
        general_args = [
            Parameter(arg.arg, ast_unparse(arg.annotation), default)
            for arg, default in zip(func.args.args, pos_defaults)
        ]
        arguments += general_args
        if vararg := func.args.vararg:
            arguments.append(
                Parameter(f"*{vararg.arg}", ast_unparse(vararg.annotation), None)
            )
        kw_only_args = [
            Parameter(
                arg.arg,
                ast_unparse(arg.annotation),
                ast_unparse(default),
            )
            for arg, default in zip(func.args.kwonlyargs, func.args.kw_defaults)
        ]
        arguments += kw_only_args
        if kwarg := func.args.kwarg:
            arguments.append(
                Parameter(f"**{kwarg.arg}", ast_unparse(kwarg.annotation), None)
            )
        # Filter out unused arguments.
        return (
            [
                argument
                for argument in arguments
                if not argument.arg_name.startswith("_")
            ]
            if self.settings.ignore_unused_arguments
            else arguments
        )

    @staticmethod
    def is_shebang_or_pragma(line: str) -> bool:
        """Check if a given line contains encoding or shebang.

        Parameters
        ----------
        line : str
            Line to check

        Returns
        -------
        bool
            Whether the given line contains encoding or shebang
        """
        shebang_regex = r"^#!(.*)"
        if re.search(shebang_regex, line) is not None:
            return True
        pragma_regex = r"^#.*coding[=:]\s*([-\w.]+)"
        return re.search(pragma_regex, line) is not None

    def get_padded_args_defaults(
        self,
        func: Union[ast.FunctionDef, ast.AsyncFunctionDef],
    ) -> list[Optional[str]]:
        """Left-Pad the general args defaults to the length of the args.

        Parameters
        ----------
        func : Union[ast.FunctionDef, ast.AsyncFunctionDef]
            Node representing a function definition

        Returns
        -------
        list[Optional[str]]
            Left padded (with `None`) list of function arguments.
        """
        pos_defaults = [ast_unparse(default) for default in func.args.defaults]
        return [None] * (len(func.args.args) - len(pos_defaults)) + pos_defaults

    def _has_excluding_decorator(
        self, node: Union[ast.FunctionDef, ast.AsyncFunctionDef]
    ) -> bool:
        """Exclude function with some decorators.

        Currently excluded:
            staticmethod
            classmethod
            property (and related)

        Parameters
        ----------
        node : Union[ast.FunctionDef, ast.AsyncFunctionDef]
            Node representing a function definition

        Returns
        -------
        bool
            Whether the function as any decorators that exclude it from
            being recognized as a standard method.
        """
        decorators = node.decorator_list
        excluded_decorators = {"staticmethod", "classmethod", "property"}
        for decorator in decorators:
            if isinstance(decorator, ast.Name) and decorator.id in excluded_decorators:
                return True
            # Handle property related decorators like in
            # @x.setter
            # def x(self, value):
            #     self._x = value  # noqa: ERA001

            # @x.deleter
            # def x(self):
            #     del self._x
            if (
                isinstance(decorator, ast.Attribute)
                and isinstance(decorator.value, ast.Name)
                and decorator.value.id == node.name
            ):
                return True
        return False

    def _check_if_node_is_self_attributes(
        self, node: ast.expr
    ) -> TypeGuard[ast.Attribute]:
        """Check whether the node represents a public attribute of self (self.abc).

        Parameters
        ----------
        node : ast.expr
            Node representing the expression to be checked.

        Returns
        -------
        TypeGuard[ast.Attribute]
            True if the node represents a public attribute of self.
        """
        return (
            isinstance(node, ast.Attribute)
            and isinstance(node.value, ast.Name)
            and node.value.id == "self"
            and not (self.settings.ignore_privates and node.attr.startswith("_"))
        )

    def _check_and_handle_assign_node(
        self, target: ast.expr, attributes: list[Parameter]
    ) -> None:
        """Check if the assignment node contains assignments to self.X.

        Add it to the list of attributes if that is the case.

        Parameters
        ----------
        target : ast.expr
            Node representing an assignment
        attributes : list[Parameter]
            List of attributes the node attribute should be added to.
        """
        if isinstance(target, (ast.Tuple, ast.List)):
            for node in target.elts:
                if self._check_if_node_is_self_attributes(node):
                    attributes.append(Parameter(node.attr, "_type_", None))
        elif self._check_if_node_is_self_attributes(target):
            attributes.append(Parameter(target.attr, "_type_", None))

    def _get_attributes_from_init(
        self, init: Union[ast.FunctionDef, ast.AsyncFunctionDef]
    ) -> list[Parameter]:
        """Iterate over body and grab every assignment `self.abc = XYZ`.

        Parameters
        ----------
        init : Union[ast.FunctionDef, ast.AsyncFunctionDef]
            Init function node to extract attributes from.

        Returns
        -------
        list[Parameter]
            List of attributes extracted from the init function.
        """
        attributes: list[Parameter] = []
        for node in init.body:
            if isinstance(node, ast.Assign):
                # Targets is a list in case of multiple assignent
                # a = b = 3  # noqa: ERA001
                for target in node.targets:
                    self._check_and_handle_assign_node(target, attributes)
            # Also handle annotated assignments
            # c: int = "Test"  # noqa: ERA001
            elif isinstance(node, ast.AnnAssign):
                self._check_and_handle_assign_node(node.target, attributes)
        return attributes

    def _get_method_signature(
        self, func: Union[ast.FunctionDef, ast.AsyncFunctionDef]
    ) -> str:
        """Remove self from signature and return the unparsed string.

        Parameters
        ----------
        func : Union[ast.FunctionDef, ast.AsyncFunctionDef]
            Node representing a function definition.

        Returns
        -------
        str
            String of the method signature with `self` removed.
        """
        arguments = func.args
        if arguments.posonlyargs:
            arguments.posonlyargs = [
                arg for arg in arguments.posonlyargs if arg.arg != "self"
            ]
        elif arguments.args:
            arguments.args = [arg for arg in arguments.args if arg.arg != "self"]
        return f"{func.name}({ast.unparse(arguments)})"

    def _get_ids_from_returns(self, values: list[ast.expr]) -> tuple[str, ...]:
        """Get the ids/names for all the expressions in the list.

        Parameters
        ----------
        values : list[ast.expr]
            List of expressions to extract the ids from.

        Returns
        -------
        tuple[str, ...]
            Tuple of ids of the original expressions.
        """
        return tuple(
            value.id
            for value in values
            # Needed again for type checker
            if isinstance(value, ast.Name)
        )



"""Integration tests of output to numpy format."""

import os
import re

import pytest

import pymend.pymend as pym
from pymend.types import FixerSettings

CURRENT_DIR = os.path.dirname(__file__)


def absdir(file: str) -> str:
    """Get absolute path for file.

    Parameters
    ----------
    file : str
        File path

    Returns
    -------
    str
        Absolute path to file
    """
    return os.path.join(CURRENT_DIR, file)


def get_expected_patch(name: str) -> str:
    """Open a patch file, and if found Pymend signature remove the 2 first lines.

    Parameters
    ----------
    name : str
        Name of the patch

    Returns
    -------
    str
        Expected patch as a string.
    """
    try:
        with open(absdir(f"refs/{name}"), encoding="utf-8") as file:
            expected_lines = file.readlines()
            if expected_lines[0].startswith("# Patch"):
                expected_lines = expected_lines[2:]
            expected = "".join(expected_lines)
    except Exception as error:  # noqa: BLE001
        pytest.fail(f'Raised exception: "{error}"')
    return expected


def remove_diff_header(diff: str) -> str:
    """Remove header differences from diff.

    Parameters
    ----------
    diff : str
        Diff file to clean.

    Returns
    -------
    str
        Cleaned diff.
    """
    return re.sub(r"(@@.+@@)|(\-\-\-.*)|(\+\+\+.*)", "", diff)


def check_expected_diff(test_name: str) -> None:
    """Check that the patch on source_file equals the expected patch."""
    expected = get_expected_patch(f"{test_name}.py.patch.numpydoc.expected")
    comment = pym.PyComment(
        absdir(f"refs/{test_name}.py"), fixer_settings=FixerSettings()
    )
    result = "".join(comment._docstring_diff())
    assert remove_diff_header(result) == remove_diff_header(expected)


class TestNumpyOutput:
    """Integration tests for numpy style output."""

    def test_positional_only_identifier(self) -> None:
        """Make sure that '/' is parsed correctly in signatures."""
        check_expected_diff("positional_only")

    def test_keyword_only_identifier(self) -> None:
        """Make sure that '*' is parsed correctly in signatures."""
        check_expected_diff("keyword_only")

    def test_returns(self) -> None:
        """Make sure single and multi return values are parsed/produced correctly."""
        check_expected_diff("returns")

    def test_star_args(self) -> None:
        """Make sure that *args are treated correctly."""
        check_expected_diff("star_args")

    def test_starstar_kwargs(self) -> None:
        """Make sure that **kwargs are treated correctly."""
        check_expected_diff("star_star_kwargs")

    def test_module_doc_dot(self) -> None:
        """Make sure missing '.' are added to the first line of module docstring."""
        check_expected_diff("module_dot_missing")

    def test_ast_ref(self) -> None:
        """Bunch of different stuff."""
        check_expected_diff("ast_ref")

    def test_yields(self) -> None:
        """Make sure yields are handled correctly from body."""
        check_expected_diff("yields")

    def test_raises(self) -> None:
        """Make sure raises are handled correctly from body."""
        check_expected_diff("raises")

    def test_skip_overload(self) -> None:
        """Function annotated with @overload should be skipped for DS creation."""
        check_expected_diff("skip_overload_decorator")

    def test_class_body(self) -> None:
        """Correctly parse and compose class from body information."""
        check_expected_diff("class_body")

    def test_quote_default(self) -> None:
        """Test that default values of triple quotes do not cause issues."""
        check_expected_diff("quote_default")

    def test_blank_lines(self) -> None:
        """Test that blank lines are set correctly."""
        expected = get_expected_patch("blank_lines.py.patch.numpydoc.expected")
        comment = pym.PyComment(
            absdir("refs/blank_lines.py"),
            fixer_settings=FixerSettings(force_params=False),
        )
        result = "".join(comment._docstring_diff())
        assert remove_diff_header(result) == remove_diff_header(expected)



"""Summarize pymend runs to users."""
from dataclasses import dataclass, field
from enum import Enum

from click import style
from typing_extensions import override

from .output import err, out


class Changed(Enum):
    """Enum for changed status."""

    NO = 0
    YES = 1


class NothingChanged(UserWarning):
    """Raised when reformatted code is the same as source."""


@dataclass
class Report:
    """Provides a reformatting counter. Can be rendered with `str(report)`."""

    check: bool = False
    diff: bool = False
    quiet: bool = False
    verbose: bool = False
    change_count: int = 0
    same_count: int = 0
    failure_count: int = 0
    issue_count: int = 0
    issues: list[str] = field(default_factory=list)

    def done(
        self, src: str, *, changed: Changed, issues: bool, issue_report: str
    ) -> None:
        """Increment the counter for successful reformatting. Write out a message.

        Parameters
        ----------
        src : str
            Source file that was successfully fixed.
        changed : Changed
            Whether the file was changed.
        issues : bool
            Whether the file had any issues.
        issue_report : str
            Issue report for the file at question.
        """
        if issues or changed == Changed.YES:
            self.issue_count += 1
            self.issues.append(issue_report)
            if changed == Changed.YES:
                reformatted = "would reformat" if self.diff else "reformatted"
                self.change_count += 1
            else:
                reformatted = "had issues"
            if self.verbose or not self.quiet:
                out(f"{reformatted} {src}")
        else:
            if self.verbose:
                msg = f"{src} already well formatted, good job."
                out(msg, bold=False)
            self.same_count += 1

    def failed(self, src: str, message: str) -> None:
        """Increment the counter for failed reformatting. Write out a message.

        Parameters
        ----------
        src : str
            File that failed to reformat.
        message : str
            Custom message to output. Should be the reason for the failure.
        """
        err(f"error: cannot format {src}: {message}")
        self.failure_count += 1

    def path_ignored(self, path: str, message: str) -> None:
        """Write out a message if a specific path was ignored.

        Parameters
        ----------
        path : str
            Path that was ignored.
        message : str
            Reason the path was ignored.
        """
        if self.verbose:
            out(f"{path} ignored: {message}", bold=False)

    @property
    def return_code(self) -> int:
        """Return the exit code that the app should use.

        This considers the current state of changed files and failures:
        - if there were any failures, return 123;
        - if any files were changed and --check is being used, return 1;
        - otherwise return 0.

        Returns
        -------
        int
            return code.
        """
        # According to http://tldp.org/LDP/abs/html/exitcodes.html starting with
        # 126 we have special return codes reserved by the shell.
        if self.failure_count:
            return 123

        if self.issue_count and self.check:
            return 1

        return 0

    @override
    def __str__(self) -> str:
        """Render a color report of the current state.

        Use `click.unstyle` to remove colors.

        Returns
        -------
        str
            Pretty string representation of the report.
        """
        if self.diff:
            reformatted = "would be reformatted"
            unchanged = "would be left unchanged"
            failed = "would fail to reformat"
        else:
            reformatted = "reformatted"
            unchanged = "left unchanged"
            failed = "failed to reformat"
        report: list[str] = []
        if self.change_count:
            s = "s" if self.change_count > 1 else ""
            report.append(
                style(f"{self.change_count} file{s} ", bold=True, fg="blue")
                + style(f"{reformatted}", bold=True)
            )
        issue_report = ""
        if self.same_count:
            s = "s" if self.same_count > 1 else ""
            report.append(style(f"{self.same_count} file{s} ", fg="blue") + unchanged)
        if self.failure_count:
            s = "s" if self.failure_count > 1 else ""
            report.append(style(f"{self.failure_count} file{s} {failed}", fg="red"))
        if self.check and self.issue_count:
            s = "s" if self.issue_count > 1 else ""
            report.append(style(f"{self.issue_count} file{s} had issues", fg="red"))
            issue_report = "\n\n" + "\n".join(
                style(msg, fg="red") for msg in self.issues
            )
        return ", ".join(report) + "." + issue_report



"""Module for defining commonly used types."""

import ast
import re
from collections.abc import Iterable, Iterator
from dataclasses import dataclass, field
from typing import Optional, Union

from typing_extensions import TypeAlias, override

import pymend.docstring_parser as dsp

from .const import DEFAULT_DESCRIPTION, DEFAULT_SUMMARY, DEFAULT_TYPE

__author__ = "J-E. Nitschke"
__copyright__ = "Copyright 2023-2023"
__licence__ = "GPL3"
__version__ = "1.0.0"
__maintainer__ = "J-E. Nitschke"


@dataclass(frozen=True)
class FixerSettings:
    """Settings to influence which sections are required and when."""

    force_params: bool = True
    force_return: bool = True
    force_raises: bool = True
    force_methods: bool = False
    force_attributes: bool = False
    force_params_min_n_params: int = 0
    force_meta_min_func_length: int = 0
    ignore_privates: bool = True
    ignore_unused_arguments: bool = True
    ignored_decorators: list[str] = field(default_factory=lambda: ["overload"])
    ignored_functions: list[str] = field(default_factory=lambda: ["main"])
    ignored_classes: list[str] = field(default_factory=list)
    force_defaults: bool = True


@dataclass
class DocstringInfo:
    """Wrapper around raw docstring."""

    name: str
    docstring: str
    lines: tuple[int, Optional[int]]
    modifier: str
    issues: list[str]

    def output_docstring(
        self,
        *,
        settings: FixerSettings,
        output_style: dsp.DocstringStyle = dsp.DocstringStyle.NUMPYDOC,
        input_style: dsp.DocstringStyle = dsp.DocstringStyle.AUTO,
    ) -> str:
        """Parse and fix input docstrings, then compose output docstring.

        Parameters
        ----------
        settings : FixerSettings
            Settings for what to fix and when.
        output_style : dsp.DocstringStyle
            Output style to use for the docstring.
            (Default value = dsp.DocstringStyle.NUMPYDOC)
        input_style : dsp.DocstringStyle
            Input style to assume for the docstring.
            (Default value = dsp.DocstringStyle.AUTO)

        Returns
        -------
        str
            String representing the updated docstring.

        Raises
        ------
        AssertionError
            If the docstring could not be parsed.
        """
        self._escape_triple_quotes()
        try:
            parsed = dsp.parse(self.docstring, style=input_style)
        except Exception as e:  # noqa: BLE001
            msg = "Failed to parse docstring with error: {e}."
            raise AssertionError(msg) from e
        self._fix_docstring(parsed, settings)
        return dsp.compose(parsed, style=output_style)

    def report_issues(self) -> tuple[int, str]:
        """Report all issues that were found in this docstring.

        Returns
        -------
        tuple[int, str]
            The number of issues found and a string representing a summary
            of those.
        """
        if not self.issues:
            return 0, ""
        return len(self.issues), f"\n{self.name}:\n" + "\n".join(self.issues)

    def _escape_triple_quotes(self) -> None:
        r"""Escape \"\"\" in the docstring."""
        if '"""' in self.docstring:
            self.issues.append("Unescaped triple quotes found.")
            self.docstring = self.docstring.replace('"""', r"\"\"\"")

    def _fix_docstring(
        self, docstring: dsp.Docstring, _settings: FixerSettings
    ) -> None:
        """Fix docstrings.

        Default are to add missing dots, blank lines and give defaults for
        descriptions and types.

        Parameters
        ----------
        docstring : dsp.Docstring
            Docstring to fix.
        settings : FixerSettings
            Settings for what to fix and when.
        """
        self._fix_backslashes()
        self._fix_short_description(docstring)
        self._fix_blank_lines(docstring)
        self._fix_descriptions(docstring)
        self._fix_types(docstring)

    def _fix_backslashes(self) -> None:
        """If there is any backslash in the docstring set it as raw."""
        if "\\" in self.docstring and "r" not in self.modifier:
            self.issues.append("Missing 'r' modifier.")
            self.modifier = "r" + self.modifier

    def _fix_short_description(self, docstring: dsp.Docstring) -> None:
        """Set default summary.

        Parameters
        ----------
        docstring : dsp.Docstring
            Docstring to set the default summary for.
        """
        cleaned_short_description = (
            docstring.short_description.strip() if docstring.short_description else ""
        )
        if (
            not cleaned_short_description
            or cleaned_short_description == DEFAULT_SUMMARY
        ):
            self.issues.append("Missing short description.")
        docstring.short_description = cleaned_short_description or DEFAULT_SUMMARY
        if not docstring.short_description.endswith("."):
            self.issues.append("Short description missing '.' at the end.")
            docstring.short_description = f"{docstring.short_description.rstrip()}."

    def _fix_blank_lines(self, docstring: dsp.Docstring) -> None:
        """Set blank lines after short and long description.

        Parameters
        ----------
        docstring : dsp.Docstring
            Docstring to fix the blank lines for.
        """
        # For parsing a blank line is associated with the description.
        if bool(docstring.blank_after_short_description) != bool(
            docstring.long_description or docstring.meta
        ):
            self.issues.append("Incorrect blank line after short description.")
        too_much = bool(
            docstring.long_description and docstring.blank_after_long_description
        ) and not bool(docstring.meta)
        missing = (
            docstring.long_description
            and not docstring.blank_after_long_description
            and docstring.meta
        )
        if too_much or missing:
            self.issues.append("Incorrect blank line after long description.")
        # Set blank line after short description if a long one follows
        # If nothing follows we do not want one and other sections bring their own.
        # For composing.
        docstring.blank_after_short_description = bool(docstring.long_description)
        # If there is a section after the long description then that already
        # introduces a newline. If not, we do not want one at all.
        docstring.blank_after_long_description = False

    def _fix_descriptions(self, docstring: dsp.Docstring) -> None:
        """Everything should have a description.

        Parameters
        ----------
        docstring : dsp.Docstring
            Docstring whose descriptions need fixing.
        """
        for ele in docstring.meta:
            # Description works a bit different for examples.
            if isinstance(ele, dsp.DocstringExample):
                continue
            if not ele.description or ele.description == DEFAULT_DESCRIPTION:
                self.issues.append(
                    f"Missing or default description `{ele.description}`."
                )
            ele.description = ele.description or DEFAULT_DESCRIPTION

    def _fix_types(self, docstring: dsp.Docstring) -> None:
        """Set empty types for parameters and returns.

        Parameters
        ----------
        docstring : dsp.Docstring
            Docstring whose type information needs fixing
        """
        for param in docstring.params:
            if param.args[0] == "method":
                continue
            if not param.type_name or param.type_name == DEFAULT_TYPE:
                self.issues.append(
                    f"Missing or default type name for parameter `{param.arg_name}`."
                )
            param.type_name = param.type_name or DEFAULT_TYPE
        for returned in docstring.many_returns:
            if not returned.type_name or returned.type_name == DEFAULT_TYPE:
                self.issues.append(
                    "Missing or default type name for return value: "
                    f" `{returned.return_name} |"
                    f" {returned.type_name} |"
                    f" {returned.description}`."
                )
            returned.type_name = returned.type_name or DEFAULT_TYPE


@dataclass
class ModuleDocstring(DocstringInfo):
    """Information about a module."""


@dataclass
class Parameter:
    """Info for parameter from signature."""

    arg_name: str
    type_name: Optional[str] = None
    default: Optional[str] = None

    def custom_hash(self) -> int:
        """Implement custom has function for uniquefying.

        Returns
        -------
        int
            Hash value of the instance.
        """
        return hash((self.arg_name, self.type_name, self.default))

    @staticmethod
    def uniquefy(lst: Iterable["Parameter"]) -> Iterator["Parameter"]:
        """Remove duplicates while keeping order.

        Parameters
        ----------
        lst : Iterable['Parameter']
            Iterable of parameters that should be uniqueified.

        Yields
        ------
        'Parameter'
            Uniqueified parameters.
        """
        seen: set[int] = set()
        for item in lst:
            if (itemhash := item.custom_hash()) not in seen:
                seen.add(itemhash)
                yield item


@dataclass
class ClassDocstring(DocstringInfo):
    """Information about a module."""

    attributes: list[Parameter]
    methods: list[str]

    @override
    def _fix_docstring(self, docstring: dsp.Docstring, settings: FixerSettings) -> None:
        """Fix docstrings.

        Additionally adjust attributes and methods from body.

        Parameters
        ----------
        docstring : dsp.Docstring
            Docstring to fix.
        settings : FixerSettings
            Settings for what to fix and when.
        """
        super()._fix_docstring(docstring, settings)
        self._adjust_attributes(docstring, settings)
        self._adjust_methods(docstring, settings)

    def _adjust_attributes(
        self, docstring: dsp.Docstring, settings: FixerSettings
    ) -> None:
        """Overwrite or create attribute docstring entries based on body.

        Create the full list if there was no original docstring.

        Do not add additional attributes and do not create the section
        if it did not exist.

        Parameters
        ----------
        docstring : dsp.Docstring
            Docstring to adjust parameters for.
        settings : FixerSettings
            Settings for what to fix and when.
        """
        # If a docstring or the section already exists we are done.
        # We already fixed empty types and descriptions in the super call.
        if self.docstring and not settings.force_attributes:
            return
        # Build dicts for faster lookup
        atts_from_doc = {
            att.arg_name: att for att in docstring.params if att.args[0] == "attribute"
        }
        atts_from_sig = {att.arg_name: att for att in self.attributes}
        for name, att_sig in atts_from_sig.items():
            # We already updated types and descriptions in the super call.
            if name in atts_from_doc:
                continue
            self.issues.append(f"Missing attribute `{att_sig.arg_name}`.")
            docstring.meta.append(
                dsp.DocstringParam(
                    args=["attribute", att_sig.arg_name],
                    description=DEFAULT_DESCRIPTION,
                    arg_name=att_sig.arg_name,
                    type_name=DEFAULT_TYPE,
                    is_optional=False,
                    default=None,
                )
            )

    def _adjust_methods(
        self, docstring: dsp.Docstring, settings: FixerSettings
    ) -> None:
        """If a new docstring is generated add a methods section.

        Create the full list if there was no original docstring.

        Do not add additional methods and do not create the section
        if it did not exist.

        Parameters
        ----------
        docstring : dsp.Docstring
            Docstring to adjust methods for.
        settings : FixerSettings
            Settings for what to fix and when.
        """
        if self.docstring and not settings.force_methods:
            return
        # Build dicts for faster lookup
        meth_from_doc = {
            meth.arg_name: meth for meth in docstring.params if meth.args[0] == "method"
        }
        for method in self.methods:
            # We already descriptions in the super call.
            if method in meth_from_doc:
                continue
            self.issues.append(f"Missing method `{method}`.")
            docstring.meta.append(
                dsp.DocstringParam(
                    args=["method", method],
                    description=DEFAULT_DESCRIPTION,
                    arg_name=method,
                    type_name=None,
                    is_optional=False,
                    default=None,
                )
            )


@dataclass
class ReturnValue:
    """Info about return value from signature."""

    type_name: Optional[str] = None


@dataclass
class FunctionSignature:
    """Information about a function signature."""

    params: list[Parameter]
    returns: ReturnValue


@dataclass
class FunctionBody:
    """Information about a function from its body."""

    raises: list[str]
    returns: set[tuple[str, ...]]
    returns_value: bool
    yields: set[tuple[str, ...]]
    yields_value: bool


@dataclass
class FunctionDocstring(DocstringInfo):
    """Information about a function from docstring."""

    signature: FunctionSignature
    body: FunctionBody
    length: int

    @override
    def _fix_docstring(self, docstring: dsp.Docstring, settings: FixerSettings) -> None:
        """Fix docstrings.

        Additionally adjust:
            parameters from function signature.
            return and yield from signature and body.
            raises from body.

        Parameters
        ----------
        docstring : dsp.Docstring
            Docstring to fix.
        settings : FixerSettings
            Settings for what to fix and when.
        """
        super()._fix_docstring(docstring, settings)
        self._adjust_parameters(docstring, settings)
        self._adjust_returns(docstring, settings)
        self._adjust_yields(docstring, settings)
        self._adjust_raises(docstring, settings)

    def _escape_default_value(self, default_value: str) -> str:
        r"""Escape the default value so that the docstring remains fully valid.

        Currently only escapes triple quotes '\"\"\"'.

        Parameters
        ----------
        default_value : str
            Value to escape.

        Returns
        -------
        str
            Optionally escaped value.
        """
        if '"""' in default_value:
            if "r" not in self.modifier:
                self.modifier = "r" + self.modifier
            return default_value.replace('"""', r"\"\"\"")
        return default_value

    def _adjust_parameters(
        self, docstring: dsp.Docstring, settings: FixerSettings
    ) -> None:
        """Overwrite or create param docstring entries based on signature.

        If an entry already exists update the type description if one exists
        in the signature. Same for the default value.

        If no entry exists then create one with name, type and default from the
        signature and place holder description.

        Parameters
        ----------
        docstring : dsp.Docstring
            Docstring to adjust parameters for.
        settings : FixerSettings
            Settings for what to fix and when.
        """
        # Build dicts for faster lookup
        params_from_doc = {param.arg_name: param for param in docstring.params}
        params_from_sig = {param.arg_name: param for param in self.signature.params}
        for name, param_sig in params_from_sig.items():
            if name in params_from_doc:
                param_doc = params_from_doc[name]
                if param_sig.type_name and param_sig.type_name != param_doc.type_name:
                    self.issues.append(
                        f"Parameter type was `{param_doc.type_name} `but signature"
                        f" has type hint `{param_sig.type_name}`."
                    )
                param_doc.type_name = param_sig.type_name or param_doc.type_name
                param_doc.is_optional = False
                if param_sig.default:
                    param_doc.default = param_sig.default
                    # param_doc.description should never be None at this point
                    # as it should have already been set by '_fix_descriptions'
                    if (
                        param_doc.description is not None
                        and "default" not in param_doc.description.lower()
                        and settings.force_defaults
                    ):
                        self.issues.append("Missing description of default value.")
                        param_doc.description += (
                            f" (Default value = "
                            f"{self._escape_default_value(param_sig.default)})"
                        )
            elif (
                settings.force_params
                and len(params_from_doc) >= settings.force_params_min_n_params
                and self.length >= settings.force_meta_min_func_length
            ):
                self.issues.append(f"Missing parameter `{name}`.")
                place_holder_description = DEFAULT_DESCRIPTION
                if param_sig.default:
                    place_holder_description += (
                        f" (Default value = "
                        f"{self._escape_default_value(param_sig.default)})"
                    )
                docstring.meta.append(
                    dsp.DocstringParam(
                        args=["param", name],
                        description=place_holder_description,
                        arg_name=name,
                        type_name=param_sig.type_name or DEFAULT_TYPE,
                        is_optional=False,
                        default=param_sig.default,
                    )
                )

    def _adjust_returns(
        self, docstring: dsp.Docstring, settings: FixerSettings
    ) -> None:
        """Overwrite or create return docstring entries based on signature.

        If no return value was parsed from the docstring:
        Add one based on the signature with a dummy description except
        if the return type was not specified or specified to be None AND there
        was an existing docstring.

        If one return value is specified overwrite the type with the signature
        if one was present there.

        If multiple were specified then leave them as is.
        They might very well be expanding on a return type like:
        Tuple[int, str, whatever]

        Parameters
        ----------
        docstring : dsp.Docstring
            Docstring to adjust return values for.
        settings : FixerSettings
            Settings for what to fix and when.
        """
        doc_returns = docstring.many_returns
        sig_return = self.signature.returns.type_name
        # If the return type is a generator extract the actual return type from that.
        if sig_return and (
            matches := (re.match(r"Generator\[(\w+), (\w+), (\w+)\]", sig_return))
        ):
            sig_return = matches[3]
        # If only one return value is specified take the type from the signature
        # as that is more likely to be correct
        if (
            not doc_returns
            and self.body.returns_value
            # If we do not want to force returns then only add new ones if
            # there was no docstring at all.
            and (
                settings.force_return
                and self.length >= settings.force_meta_min_func_length
                or not self.docstring
            )
        ):
            self.issues.append("Missing return value.")
            docstring.meta.append(
                dsp.DocstringReturns(
                    args=["returns"],
                    description=DEFAULT_DESCRIPTION,
                    type_name=sig_return or DEFAULT_TYPE,
                    is_generator=False,
                    return_name=None,
                )
            )
        # If there is only one return value specified and we do not
        # yield anything then correct it with the actual return value.
        elif len(doc_returns) == 1 and not self.body.yields_value:
            doc_return = doc_returns[0]
            if sig_return and doc_return.type_name != sig_return:
                self.issues.append(
                    f"Return type was `{doc_return.type_name}` but"
                    f" signature has type hint `{sig_return}`."
                )
            doc_return.type_name = sig_return or doc_return.type_name
        # If we have multiple return values specified
        # and we have only extracted one set of return values from the body.
        # then update the multiple return values with the names from
        # the actual return values.
        elif len(doc_returns) > 1 and len(self.body.returns) == 1:
            doc_names = {returned.return_name for returned in doc_returns}
            for body_name in next(iter(self.body.returns)):
                if body_name not in doc_names:
                    self.issues.append(
                        f"Missing return value in multi return statement `{body_name}`."
                    )
                    docstring.meta.append(
                        dsp.DocstringReturns(
                            args=["returns"],
                            description=DEFAULT_DESCRIPTION,
                            type_name=DEFAULT_TYPE,
                            is_generator=False,
                            return_name=body_name,
                        )
                    )

    def _adjust_yields(self, docstring: dsp.Docstring, settings: FixerSettings) -> None:
        """See _adjust_returns.

        Only difference is that the signature return type is not added
        to the docstring since it is a bit more complicated for generators.

        Parameters
        ----------
        docstring : dsp.Docstring
            Docstring to adjust yields for.
        settings : FixerSettings
            Settings for what to fix and when.
        """
        doc_yields = docstring.many_yields
        sig_return = self.signature.returns.type_name
        # Extract actual return type from Iterators and Generators.
        if sig_return and (
            matches := (
                re.match(r"(?:Iterable|Iterator)\[([^\]]+)\]", sig_return)
                or re.match(r"Generator\[(\w+), (\w+), (\w+)\]", sig_return)
            )
        ):
            sig_return = matches[1]
        else:
            sig_return = None
        # If only one return value is specified take the type from the signature
        # as that is more likely to be correct
        if not doc_yields and self.body.yields_value and settings.force_return:
            self.issues.append("Missing yielded value.")
            docstring.meta.append(
                dsp.DocstringYields(
                    args=["yields"],
                    description=DEFAULT_DESCRIPTION,
                    type_name=sig_return or DEFAULT_TYPE,
                    is_generator=True,
                    yield_name=None,
                )
            )
        elif len(doc_yields) == 1:
            doc_yields = doc_yields[0]
            if sig_return and doc_yields.type_name != sig_return:
                self.issues.append(
                    f"Yield type was `{doc_yields.type_name}` but"
                    f" signature has type hint `{sig_return}`."
                )
            doc_yields.type_name = sig_return or doc_yields.type_name
        elif len(doc_yields) > 1 and len(self.body.yields) == 1:
            doc_names = {yielded.yield_name for yielded in doc_yields}
            for body_name in next(iter(self.body.yields)):
                if body_name not in doc_names:
                    self.issues.append(
                        f"Missing return value in multi return statement `{body_name}`."
                    )
                    docstring.meta.append(
                        dsp.DocstringYields(
                            args=["yields"],
                            description=DEFAULT_DESCRIPTION,
                            type_name=DEFAULT_TYPE,
                            is_generator=True,
                            yield_name=body_name,
                        )
                    )

    def _adjust_raises(self, docstring: dsp.Docstring, settings: FixerSettings) -> None:
        """Adjust raises section based on parsed body.

        Parameters
        ----------
        docstring : dsp.Docstring
            Docstring to adjust raises section for.
        settings : FixerSettings
            Settings for what to fix and when.
        """
        if self.docstring and not settings.force_raises:
            return
        # Only consider those raises that are not already raised in the body.
        # We are potentially raising the same time of exception multiple times.
        # Only remove the first of each type per one encountered in the docstring.
        raised_in_body = self.body.raises.copy()
        for raised in docstring.raises:
            if raised.type_name in raised_in_body:
                raised_in_body.remove(raised.type_name)
            # If this specific Error is not in the body but the body contains
            # unknown exceptions then remove one of those instead.
            # For example when exception stored in variable and raised later.
            # We want people to be able to specific them by name and not have
            # pyment constantly forced unnamed raises on them.
            elif "" in raised_in_body:
                raised_in_body.remove("")
        for missing_raise in raised_in_body:
            self.issues.append(f"Missing raised exception `{missing_raise}`.")
            docstring.meta.append(
                dsp.DocstringRaises(
                    args=["raises", missing_raise],
                    description=DEFAULT_DESCRIPTION,
                    type_name=missing_raise,
                )
            )


ElementDocstring: TypeAlias = Union[ModuleDocstring, ClassDocstring, FunctionDocstring]
DefinitionNodes: TypeAlias = Union[ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef]
NodeOfInterest: TypeAlias = Union[DefinitionNodes, ast.Module]



"""File handling for pymend."""

import sys
from collections.abc import Sequence
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional

if sys.version_info >= (3, 11):
    try:
        import tomllib
    except ImportError:
        # Help users on older alphas
        if not TYPE_CHECKING:
            import tomli as tomllib
        else:
            raise
else:
    import tomli as tomllib


@lru_cache
def find_project_root(srcs: Sequence[str]) -> tuple[Path, str]:
    """Return a directory containing .git, .hg, or pyproject.toml.

    That directory will be a common parent of all files and directories
    passed in `srcs`.

    If no directory in the tree contains a marker that would specify it's the
    project root, the root of the file system is returned.

    Returns a two-tuple with the first element as the project root path and
    the second element as a string describing the method by which the
    project root was discovered.

    Parameters
    ----------
    srcs : Sequence[str]
        Source files that will be considered by pymend.

    Returns
    -------
    directory : Path
        Projects root path
    method : str
        Method by which the root path was determined.
    """
    if not srcs:
        srcs = [str(Path.cwd().resolve())]

    path_srcs = [Path(Path.cwd(), src).resolve() for src in srcs]

    # A list of lists of parents for each 'src'. 'src' is included as a
    # "parent" of itself if it is a directory
    src_parents = [
        list(path.parents) + ([path] if path.is_dir() else []) for path in path_srcs
    ]

    intersection: set[Path] = set[Path].intersection(
        *(set(parents) for parents in src_parents)
    )

    common_base = max(
        intersection,
        key=lambda path: path.parts,
    )

    # Directory will always be set in the loop.
    # This is just for pylint.
    directory = Path("")
    for directory in (common_base, *common_base.parents):
        if (directory / ".git").exists():
            return directory, ".git directory"

        if (directory / ".hg").is_dir():
            return directory, ".hg directory"

        if (directory / "pyproject.toml").is_file():
            return directory, "pyproject.toml"

    return directory, "file system root"


def find_pyproject_toml(path_search_start: tuple[str, ...]) -> Optional[str]:
    """Find the absolute filepath to a pyproject.toml if it exists.

    Parameters
    ----------
    path_search_start : tuple[str, ...]
        Tuple of paths to consider in the search for pyproject.toml

    Returns
    -------
    Optional[str]
        Path to pypyproject.toml or None if it could not be found.
    """
    path_project_root, _ = find_project_root(path_search_start)
    path_pyproject_toml = path_project_root / "pyproject.toml"
    if path_pyproject_toml.is_file():
        return str(path_pyproject_toml)

    return None


def parse_pyproject_toml(path_config: str) -> dict[str, Any]:
    """Parse a pyproject toml file, pulling out relevant parts for pymend.

    If parsing fails, will raise a tomllib.TOMLDecodeError.

    Parameters
    ----------
    path_config : str
        Path to the pyproject.toml file.

    Returns
    -------
    dict[str, Any]
        Configuration dictionary parsed from pyproject.toml
    """
    with open(path_config, "rb") as f:
        pyproject_toml: dict[str, Any] = tomllib.load(f)
    config: dict[str, Any] = pyproject_toml.get("tool", {}).get("pymend", {})
    return {k.replace("--", "").replace("-", "_"): v for k, v in config.items()}



def func1():
    pass


def func2(param1):
    pass


def func3(param1, param2):
    pass


def func3(param1, param2, param3):
    pass


def func4(param1=123):
    pass


def func5(param1, param2=123):
    pass


def func6(param1, param2, param3=123):
    pass


def func7(param1, param2=None, param3=123):
    pass


def func8(param1=123, param2=+456, param3="!:@"):
    pass


def func9(param1=func1(), param2="stuff"):
    pass


def func10(param1=func1(), param2="stuff"):  # comments with (parentheses):
    pass



"""Unit tests for pymend.file_parser.py."""

import ast

from pymend.file_parser import AstAnalyzer
from pymend.types import FixerSettings, Parameter


class TestAstAnalyzer:
    """Test ast analyzer."""

    def test_handle_class_body(self) -> None:
        """Handle class body parsing."""
        class_definition = '''\
class C:
    def __init__(self):
        self._x = None
        self.test1 = "test"
        self.test2: Optional[int] = None
        self.test1 = "a"
        self.test3 = self.test4 = None
        self.test5, self.test6 = 1, 2

    @property
    def x(self) -> str | None:
        """I'm the 'x' property."""
        return self._x

    @x.setter
    def x(self, value):
        self._x = value

    @staticmethod
    def a(self, a):
        pass

    @classmethod
    def b(self, b):
        pass

    def c(self, c):
        pass
'''
        class_node = ast.parse(class_definition).body[0]
        analyzer = AstAnalyzer(class_definition, settings=FixerSettings())

        attributes, methods = analyzer.handle_class_body(class_node)

        expected_attributes = [
            Parameter("test1", "_type_", None),
            Parameter("test2", "_type_", None),
            Parameter("test3", "_type_", None),
            Parameter("test4", "_type_", None),
            Parameter("test5", "_type_", None),
            Parameter("test6", "_type_", None),
            Parameter("x", "str | None", None),
        ]
        expected_methods = ["c(c)"]
        assert attributes == expected_attributes
        assert methods == expected_methods



"""Google-style docstring parsing."""

import inspect
import re
from collections import OrderedDict
from collections.abc import Mapping, Sequence
from enum import IntEnum
from typing import NamedTuple, Optional

from .common import (
    EXAMPLES_KEYWORDS,
    PARAM_KEYWORDS,
    RAISES_KEYWORDS,
    RETURNS_KEYWORDS,
    YIELDS_KEYWORDS,
    Docstring,
    DocstringExample,
    DocstringMeta,
    DocstringParam,
    DocstringRaises,
    DocstringReturns,
    DocstringStyle,
    DocstringYields,
    MainSections,
    ParseError,
    RenderingStyle,
    append_description,
    split_description,
)


class SectionType(IntEnum):
    """Types of sections."""

    SINGULAR = 0
    """For sections like examples."""

    MULTIPLE = 1
    """For sections like params."""

    SINGULAR_OR_MULTIPLE = 2
    """For sections like returns or yields."""


class Section(NamedTuple):
    """A docstring section."""

    title: str
    key: str
    type_info: SectionType


GOOGLE_TYPED_ARG_REGEX = re.compile(r"\s*(.+?)\s*\(\s*(.*[^\s]+)\s*\)")
GOOGLE_ARG_DESC_REGEX = re.compile(r".*\. Defaults to (.+)\.")
MULTIPLE_PATTERN = re.compile(
    r"(\s*[^:\s]+:)|([^:]*\]:.*)|(\s*[^\s|][^\|]*[^\s|](\s*\|\s*[^\s|][^\|]*[^\s|])+:)"
)

DEFAULT_SECTIONS = [
    Section("Arguments", "param", SectionType.MULTIPLE),
    Section("Args", "param", SectionType.MULTIPLE),
    Section("Parameters", "param", SectionType.MULTIPLE),
    Section("Params", "param", SectionType.MULTIPLE),
    Section("Raises", "raises", SectionType.MULTIPLE),
    Section("Exceptions", "raises", SectionType.MULTIPLE),
    Section("Except", "raises", SectionType.MULTIPLE),
    Section("Attributes", "attribute", SectionType.MULTIPLE),
    Section("Example", "examples", SectionType.SINGULAR),
    Section("Examples", "examples", SectionType.SINGULAR),
    Section("Returns", "returns", SectionType.SINGULAR_OR_MULTIPLE),
    Section("Yields", "yields", SectionType.SINGULAR_OR_MULTIPLE),
]


class GoogleParser:
    """Parser for Google-style docstrings."""

    def __init__(
        self, sections: Optional[list[Section]] = None, *, title_colon: bool = True
    ) -> None:
        """Set up sections.

        Parameters
        ----------
        sections : Optional[list[Section]]
            Recognized sections or None to defaults.
        title_colon : bool
            Require colon after section title. (Default value = True)
        """
        if not sections:
            sections = DEFAULT_SECTIONS
        self.sections = {s.title: s for s in sections}
        self.title_colon = title_colon
        self._setup()

    def _setup(self) -> None:
        """Set up parser with the colon type and title regex."""
        colon = ":" if self.title_colon else ""
        self.titles_re = re.compile(
            "^("
            + "|".join(f"({t})" for t in self.sections)
            + ")"
            + colon
            + "[ \t\r\f\v]*$",
            flags=re.M,
        )

    def _build_meta(self, text: str, title: str) -> DocstringMeta:
        """Build docstring element.

        Parameters
        ----------
        text : str
            docstring element text
        title : str
            title of section containing element

        Returns
        -------
        DocstringMeta
            docstring meta element

        Raises
        ------
        ParseError
            If the text lacks a colon ':'
        """
        section = self.sections[title]

        if (
            section.type_info == SectionType.SINGULAR_OR_MULTIPLE
            and not MULTIPLE_PATTERN.match(text)
        ) or section.type_info == SectionType.SINGULAR:
            return self._build_single_meta(section, text)

        if ":" not in text:
            msg = f"Expected a colon in {text!r} for title {title}."
            raise ParseError(msg)

        # Split spec and description
        before, desc = text.split(":", 1)
        if desc:
            desc = desc[1:] if desc[0] == " " else desc
            if "\n" in desc:
                first_line, rest = desc.split("\n", 1)
                desc = first_line + "\n" + inspect.cleandoc(rest)
            desc = desc.strip("\n")

        return self._build_multi_meta(section, before, desc)

    @staticmethod
    def _build_single_meta(section: Section, desc: str) -> DocstringMeta:
        """Build docstring element for single line sections.

        Parameters
        ----------
        section : Section
            The section that is being processed.
        desc : str
            docstring element text

        Returns
        -------
        DocstringMeta
            Docstring meta wrapper.

        Raises
        ------
        ParseError
            If the section represents a parameter section.
            In that case we would not expect to be in the single line function.
        """
        if section.key in RETURNS_KEYWORDS:
            return DocstringReturns(
                args=[section.key],
                description=desc,
                type_name=None,
                is_generator=False,
            )
        if section.key in YIELDS_KEYWORDS:
            return DocstringYields(
                args=[section.key],
                description=desc,
                type_name=None,
                is_generator=True,
            )
        if section.key in RAISES_KEYWORDS:
            return DocstringRaises(args=[section.key], description=desc, type_name=None)
        if section.key in EXAMPLES_KEYWORDS:
            return DocstringExample(args=[section.key], snippet=None, description=desc)
        if section.key in PARAM_KEYWORDS:
            msg = "Expected parameter name."
            raise ParseError(msg)
        return DocstringMeta(args=[section.key], description=desc)

    @staticmethod
    def _build_multi_meta(section: Section, before: str, desc: str) -> DocstringMeta:
        """Build docstring element for multi line sections.

        Parameters
        ----------
        section : Section
            The section that is being processed.
        before : str
            Part before the first colon in the docstring text.
        desc : str
            Rest of the description.

        Returns
        -------
        DocstringMeta
            Docstring meta wrapper around.
        """
        if section.key in PARAM_KEYWORDS:
            match = GOOGLE_TYPED_ARG_REGEX.match(before)
            if match:
                arg_name, type_name = match.group(1, 2)
                if type_name.endswith(", optional"):
                    is_optional = True
                    type_name = type_name[:-10]
                elif type_name.endswith("?"):
                    is_optional = True
                    type_name = type_name[:-1]
                else:
                    is_optional = False
            else:
                arg_name, type_name = before, None
                is_optional = None

            match = GOOGLE_ARG_DESC_REGEX.match(desc)
            default = match.group(1) if match else None

            return DocstringParam(
                args=[section.key, before],
                description=desc,
                arg_name=arg_name,
                type_name=type_name,
                is_optional=is_optional,
                default=default,
            )
        if section.key in RETURNS_KEYWORDS:
            return DocstringReturns(
                args=[section.key, before],
                description=desc,
                type_name=before,
                is_generator=False,
            )
        if section.key in YIELDS_KEYWORDS:
            return DocstringYields(
                args=[section.key, before],
                description=desc,
                type_name=before,
                is_generator=True,
            )
        if section.key in RAISES_KEYWORDS:
            return DocstringRaises(
                args=[section.key, before], description=desc, type_name=before
            )
        return DocstringMeta(args=[section.key, before], description=desc)

    def add_section(self, section: Section) -> None:
        """Add or replace a section.

        Parameters
        ----------
        section : Section
            The new section.
        """
        self.sections[section.title] = section
        self._setup()

    def _split_sections(self, meta_chunk: str) -> Mapping[str, str]:
        """Split the cunk into sections as determined by the titles..

        Parameters
        ----------
        meta_chunk : str
            Part of the docstring NOT holding the description.

        Returns
        -------
        Mapping[str, str]
            Mapping between sectrion title and part of the docstring that deals with it.
        """
        chunks: Mapping[str, str] = OrderedDict()

        matches = list(self.titles_re.finditer(meta_chunk))
        if not matches:
            return chunks
        splits = [
            (matches[j].end(), matches[j + 1].start()) for j in range(len(matches) - 1)
        ]
        splits.append((matches[-1].end(), len(meta_chunk)))

        for j, (start, end) in enumerate(splits):
            title = matches[j].group(1)
            if title not in self.sections:
                continue

            # Clear Any Unknown Meta
            # Ref: https://github.com/rr-/docstring_parser/issues/29
            meta_details = meta_chunk[start:end]
            unknown_meta = re.search(r"\n\S", meta_details)
            if unknown_meta is not None:
                meta_details = meta_details[: unknown_meta.start()]

            chunks[title] = meta_details.strip("\n")
        return chunks

    def _determine_indent(self, chunk: str) -> str:
        """Determine indent.

        Parameters
        ----------
        chunk : str
            Chunk to determine the indent for.

        Returns
        -------
        str
            String representing the indent.

        Raises
        ------
        ParseError
            If no indent could be determined.
        """
        indent_match = re.search(r"^\s*", chunk)
        if not indent_match:
            msg = f"Can't infer indent from '{chunk}'"
            raise ParseError(msg)
        return indent_match.group()

    def _get_chunks(self, text: str) -> tuple[str, str]:
        """Split docstring into description and meta part.

        Parameters
        ----------
        text : str
            Docstring text to split.

        Returns
        -------
        tuple[str, str]
            Docstring representing the description and the rest.
        """
        if match := self.titles_re.search(text):
            return text[: match.start()], text[match.start() :]
        return text, ""

    def parse(self, text: Optional[str]) -> Docstring:
        """Parse the Google-style docstring into its components.

        Parameters
        ----------
        text : Optional[str]
            docstring text

        Returns
        -------
        Docstring
            parsed docstring

        Raises
        ------
        ParseError
            If no specification could be found for a title, chunk pair.
        """
        ret = Docstring(style=DocstringStyle.GOOGLE)
        if not text:
            return ret

        # Clean according to PEP-0257
        text = inspect.cleandoc(text)

        desc_chunk, meta_chunk = self._get_chunks(text)

        # Break description into short and long parts
        split_description(ret, desc_chunk)

        # Split by sections determined by titles
        chunks = self._split_sections(meta_chunk)

        if not chunks:
            return ret

        # Add elements from each chunk
        for title, chunk in chunks.items():
            # Determine indent
            indent = self._determine_indent(chunk)

            # Check for singular elements
            if self.sections[title].type_info in [
                SectionType.SINGULAR,
                SectionType.SINGULAR_OR_MULTIPLE,
            ]:
                part = inspect.cleandoc(chunk)
                ret.meta.append(self._build_meta(part, title))
                continue

            # Split based on lines which have exactly that indent
            c_matches = list(re.finditer(rf"^{indent}(?=\S)", chunk, flags=re.M))
            if not c_matches:
                msg = f'No specification for "{title}": "{chunk}"'
                raise ParseError(msg)
            c_splits = [
                (c_cur.end(), c_next.start())
                for c_cur, c_next in zip(c_matches, c_matches[1:])
            ]
            c_splits.append((c_matches[-1].end(), len(chunk)))
            for start, end in c_splits:
                part = chunk[start:end].strip("\n")
                ret.meta.append(self._build_meta(part, title))

        return ret


def parse(text: Optional[str]) -> Docstring:
    """Parse the Google-style docstring into its components.

    Parameters
    ----------
    text : Optional[str]
        docstring text

    Returns
    -------
    Docstring
        parsed docstring
    """
    return GoogleParser().parse(text)


def compose(  # noqa: PLR0915
    docstring: Docstring,
    rendering_style: RenderingStyle = RenderingStyle.COMPACT,
    indent: str = "    ",
) -> str:
    """Render a parsed docstring into docstring text.

    Parameters
    ----------
    docstring : Docstring
        parsed docstring representation
    rendering_style : RenderingStyle
        the style to render docstrings (Default value = RenderingStyle.COMPACT)
    indent : str
        the characters used as indentation in the
        docstring string (Default value = '    ')

    Returns
    -------
    str
        docstring text
    """

    def process_one(one: MainSections) -> None:
        """Build the output text for one entry in a section.

        Parameters
        ----------
        one : MainSections
            Docstring for which to build the raw text.
        """
        head = ""

        if isinstance(one, DocstringParam):
            head += one.arg_name or ""
        elif isinstance(one, DocstringReturns):
            head += one.return_name or ""
        elif isinstance(one, DocstringYields):
            head += one.yield_name or ""

        if isinstance(one, DocstringParam) and one.is_optional:
            optional = (
                "?" if rendering_style == RenderingStyle.COMPACT else ", optional"
            )
        else:
            optional = ""

        if one.type_name and head:
            head += f" ({one.type_name}{optional}):"
        elif one.type_name:
            head += f"{one.type_name}{optional}:"
        else:
            head += ":"
        head = indent + head

        if one.description and rendering_style == RenderingStyle.EXPANDED:
            body = f"\n{indent}{indent}".join([head, *one.description.splitlines()])
            parts.append(body)
        elif one.description:
            (first, *rest) = one.description.splitlines()
            body = f"\n{indent}{indent}".join([f"{head} {first}", *rest])
            parts.append(body)
        else:
            parts.append(head)

    def process_sect(name: str, args: Sequence[MainSections]) -> None:
        """Build the output for a docstring section.

        Parameters
        ----------
        name : str
            Section for which to build the output.
        args : Sequence[MainSections]
            List of individual elements of that section.
        """
        if args:
            parts.append(name)
            for arg in args:
                process_one(arg)
            parts.append("")

    parts: list[str] = []
    append_description(docstring, parts)

    process_sect("Args:", [p for p in docstring.params or [] if p.args[0] == "param"])

    process_sect(
        "Attributes:",
        [p for p in docstring.params or [] if p.args[0] == "attribute"],
    )

    process_sect(
        "Returns:",
        docstring.many_returns,
    )

    process_sect("Yields:", docstring.many_yields)

    process_sect("Raises:", docstring.raises or [])

    if docstring.returns and not docstring.many_returns:
        ret = docstring.returns
        parts.append("Yields:" if ret else "Returns:")
        parts.append("-" * len(parts[-1]))
        process_one(ret)

    for meta in docstring.meta:
        if isinstance(
            meta, (DocstringParam, DocstringReturns, DocstringRaises, DocstringYields)
        ):
            continue  # Already handled
        parts.append(meta.args[0].replace("_", "").title() + ":")
        if meta.description:
            lines = [indent + line for line in meta.description.splitlines()]
            parts.append("\n".join(lines))
        parts.append("")

    while parts and not parts[-1]:
        parts.pop()

    return "\n".join(parts)



# -*- coding: utf-8 -*-
#!/usr/bin/env python
"""Some module docstring
a
bit
longer"""


def func1():
    pass


def func2():
    """SOme docstring"""
    pass


def func3():
    return 1 + 1


def func_with_params(pos, a: "annotation", b: int, c: int, *args: list, d: int = 5, e="test", **kwargs: dict) -> str:
    """_summary_.

    Parameters
    ----------
    a : _type_
        _description_
    b : int
        _description_
    c : int
        _description_ (Default value = 2)
    *args : _type_
        _description_
    d : int
        _description_ (Default value = 5)
    e : _type_
        _description_ (Default value = "test")
    **kwargs : _type_
        _description_

    Returns
    -------
    str
        _description_
    """
    pass


def func2_with_params(a: str, /, b: int, c: int = 3, d: int = 5, e="test", **kwargs) -> str:
    """Short string"""


class SomeTestClass:
    """Some docstring for that class."""

    def __init__(self, one_param=2):
        pass

SomeTestClass()



"""Main pymend module."""

from .pymend import PyComment, __author__, __copyright__, __licence__, __version__

name = "pymend"  # pylint: disable=invalid-name

__all__ = ["PyComment", "__version__", "__copyright__", "__author__", "__licence__"]



#!/usr/bin/python
"""Command line interface for pymend."""

import platform
import re
import traceback
from re import Pattern
from typing import Any, Optional, Union

import click
from click.core import ParameterSource

import pymend.docstring_parser as dsp
from pymend import PyComment, __version__

from .const import DEFAULT_EXCLUDES
from .files import find_pyproject_toml, parse_pyproject_toml
from .output import out
from .report import Report
from .types import FixerSettings

STRING_TO_STYLE = {
    "rest": dsp.DocstringStyle.REST,
    "javadoc": dsp.DocstringStyle.EPYDOC,
    "numpydoc": dsp.DocstringStyle.NUMPYDOC,
    "google": dsp.DocstringStyle.GOOGLE,
}


def path_is_excluded(
    normalized_path: str,
    pattern: Optional[Pattern[str]],
) -> bool:
    """Check if a path is excluded because it matches and exclusion regex.

    Parameters
    ----------
    normalized_path : str
        Normalized path to check
    pattern : Optional[Pattern[str]]
        Optionally a regex pattern to check against

    Returns
    -------
    bool
        True if the path is excluded by the regex.
    """
    match = pattern.search(normalized_path) if pattern else None
    return bool(match and match.group(0))


def style_option_callback(
    _c: click.Context, _p: Union[click.Option, click.Parameter], style: str
) -> dsp.DocstringStyle:
    """Compute the output style from a --output_stye flag.

    Parameters
    ----------
    style : str
        String representation of the style to use.

    Returns
    -------
    dsp.DocstringStyle
        Style to use.
    """
    if style in STRING_TO_STYLE:
        return STRING_TO_STYLE[style]
    return dsp.DocstringStyle.AUTO


def re_compile_maybe_verbose(regex: str) -> Pattern[str]:
    """Compile a regular expression string in `regex`.

    If it contains newlines, use verbose mode.

    Parameters
    ----------
    regex : str
        Regex to compile.

    Returns
    -------
    Pattern[str]
        Compiled regex.
    """
    if "\n" in regex:
        regex = "(?x)" + regex
    compiled: Pattern[str] = re.compile(regex)
    return compiled


def validate_regex(
    _ctx: click.Context,
    _param: click.Parameter,
    value: Optional[str],
) -> Optional[Pattern[str]]:
    """Validate the regex from command line.

    Parameters
    ----------
    value : Optional[str]
        Regex pattern to validate.

    Returns
    -------
    Optional[Pattern[str]]
        Compiled regex pattern or None if the input was None.

    Raises
    ------
    click.BadParameter
        If the value is not a valid regex.
    """
    try:
        return re_compile_maybe_verbose(value) if value is not None else None
    except re.error as e:
        msg = f"Not a valid regular expression: {e}"
        raise click.BadParameter(msg) from None


def run(
    files: tuple[str, ...],
    *,
    overwrite: bool = False,
    output_style: dsp.DocstringStyle = dsp.DocstringStyle.NUMPYDOC,
    input_style: dsp.DocstringStyle = dsp.DocstringStyle.AUTO,
    exclude: Pattern[str],
    extend_exclude: Optional[Pattern[str]],
    report: Report,
    fixer_settings: FixerSettings,
) -> None:
    r"""Run pymend over the list of files..

    Parameters
    ----------
    files : tuple[str, ...]
        List of files to analyze and fix.
    overwrite : bool
        Whether to overwrite the source file directly instead of creating
        a patch. (Default value = False)
    output_style : dsp.DocstringStyle
        Output style to use for the modified docstrings.
        (Default value = dsp.DocstringStyle.NUMPYDOC)
    input_style : dsp.DocstringStyle
        Input docstring style.
        Auto means that the style is detected automatically. Can cause issues when
        styles are mixed in examples or descriptions."
        (Default value = dsp.DocstringStyle.AUTO)
    exclude : Pattern[str]
        Optional regex pattern to use to exclude files from reformatting.
    extend_exclude : Optional[Pattern[str]]
        Additional regexes to add onto the exclude pattern.
        Useful if one just wants to add some to the existing default.
    report : Report
        Reporter for pretty communication with the user.
    fixer_settings : FixerSettings
        Settings for which fixes should be performed.

    Raises
    ------
    AssertionError
        If the input and output lines are identical but pyment reports
        some elements to have changed.
    """
    for file in files:
        if path_is_excluded(file, exclude):
            report.path_ignored(file, "matches the --exclude regular expression")
            continue
        if path_is_excluded(file, extend_exclude):
            report.path_ignored(file, "matches the --extend-exclude regular expression")
            continue
        try:
            comment = PyComment(
                file,
                output_style=output_style,
                input_style=input_style,
                fixer_settings=fixer_settings,
            )
            n_issues, issue_report = comment.report_issues()
            # Not using ternary when the calls have side effects
            if overwrite:  # noqa: SIM108
                changed = comment.output_fix()
            else:
                changed = comment.output_patch()
            report.done(
                file, changed=changed, issues=bool(n_issues), issue_report=issue_report
            )
        except Exception as exc:  # noqa: BLE001
            if report.verbose:
                traceback.print_exc()
            report.failed(file, str(exc))


def read_pyproject_toml(
    ctx: click.Context, _param: click.Parameter, value: Optional[str]
) -> Optional[str]:
    """Inject Black configuration from "pyproject.toml" into defaults in `ctx`.

    Returns the path to a successfully found and read configuration file, None
    otherwise.

    Parameters
    ----------
    ctx : click.Context
        Context containing preexisting default values.
    value : Optional[str]
        Optionally path to the config file.

    Returns
    -------
    Optional[str]
        Path to the config file if one was found or specified.

    Raises
    ------
    click.FileError
        If there was a problem reading the configuration file.
    click.BadOptionUsage
        If the value passed for `exclude` was not a string.
    click.BadOptionUsage
        If the value passed for `extended_exclude` was not a string.
    """
    if not value:
        value = find_pyproject_toml(ctx.params.get("src", ()))
        if value is None:
            return None

    try:
        config = parse_pyproject_toml(value)
    except (OSError, ValueError) as e:
        raise click.FileError(
            filename=value, hint=f"Error reading configuration file: {e}"
        ) from None

    if not config:
        return None
    # Sanitize the values to be Click friendly. For more information please see:
    # https://github.com/psf/black/issues/1458
    # https://github.com/pallets/click/issues/1567
    config: dict[str, Any] = {
        k: str(v) if not isinstance(v, (list, dict)) else v for k, v in config.items()
    }

    exclude = config.get("exclude")
    if exclude is not None and not isinstance(exclude, str):
        raise click.BadOptionUsage("exclude", "Config key exclude must be a string")

    extend_exclude = config.get("extend_exclude")
    if extend_exclude is not None and not isinstance(extend_exclude, str):
        raise click.BadOptionUsage(
            "extend-exclude", "Config key extend-exclude must be a string"
        )

    default_map: dict[str, Any] = {}
    if ctx.default_map:
        default_map.update(ctx.default_map)
    default_map.update(config)

    ctx.default_map = default_map
    return value


@click.command(
    context_settings={"help_option_names": ["-h", "--help"]},
    help="Create, update or convert docstrings.",
)
@click.option(
    "--write/--diff",
    is_flag=True,
    default=False,
    help="Directly overwrite the source files instead of just producing a patch.",
)
@click.option(
    "-o",
    "--output-style",
    type=click.Choice(list(STRING_TO_STYLE)),
    callback=style_option_callback,
    multiple=False,
    default="numpydoc",
    help=("Output docstring style."),
)
@click.option(
    "-i",
    "--input-style",
    type=click.Choice([*list(STRING_TO_STYLE), "auto"]),
    callback=style_option_callback,
    multiple=False,
    default="auto",
    help=(
        "Input docstring style."
        " Auto means that the style is detected automatically. Can cause issues when"
        " styles are mixed in examples or descriptions."
    ),
)
@click.option(
    "--check",
    is_flag=True,
    help=(
        "Perform check if file is properly docstringed."
        " Also reports negatively on pymend defaults."
        " Return code 0 means everything was perfect."
        " Return code 1 means some files would has issues."
        " Return code 123 means there was an internal error."
    ),
)
@click.option(
    "--exclude",
    type=str,
    callback=validate_regex,
    help=(
        "A regular expression that matches files and directories that should be"
        " excluded. An empty value means no paths are excluded."
        " Use forward slashes for directories on all platforms (Windows, too)."
        f"[default: {DEFAULT_EXCLUDES}]"
    ),
    show_default=False,
)
@click.option(
    "--extend-exclude",
    type=str,
    callback=validate_regex,
    help=(
        "Like --exclude, but adds additional files and directories on top of the"
        " excluded ones. (Useful if you simply want to add to the default)"
    ),
)
@click.option(
    "--force-params/--unforce-params",
    type=bool,
    is_flag=True,
    default=True,
    help="Whether to force a parameter section even if"
    " there is already an existing docstring. "
    "If set will also fill force the parameters section to name every parameter.",
)
@click.option(
    "--force-params-min-n-params",
    type=int,
    default=0,
    help="Minimum number of arguments detected in the signature "
    "to actually enforce parameters."
    " If less than the specified numbers of arguments are"
    " detected then a parameters section is only build for new docstrings."
    " No new sections are created for existing docstrings and existing sections"
    " are not extended. Only has an effect with --force-params set to true.",
)
@click.option(
    "--force-meta-min-func-length",
    type=int,
    default=0,
    help="Minimum number statements in the function body "
    "to actually enforce parameters."
    " If less than the specified numbers of arguments are"
    " detected then a parameters section is only build for new docstrings."
    " No new sections are created for existing docstrings and existing sections"
    " are not extended. Only has an effect with"
    " `--force-params` or `--force-return` set to true."
    " Currently this only counts the length of the body of the ast node."
    " So nested statements are not taken into account (yet).",
)
@click.option(
    "--force-return/--unforce-return",
    type=bool,
    is_flag=True,
    default=True,
    help="Whether to force a return/yield section even if"
    " there is already an existing docstring. "
    "Will only actually force return/yield sections"
    " if any value return or yield is found in the body.",
)
@click.option(
    "--force-raises/--unforce-raises",
    type=bool,
    is_flag=True,
    default=True,
    help="Whether to force a raises section even if"
    " there is already an existing docstring."
    " Will only actually force the section if any raises were detected in the body."
    " However, if set it will force on entry in the section per raise detected.",
)
@click.option(
    "--force-methods/--unforce-methods",
    type=bool,
    is_flag=True,
    default=False,
    help="Whether to force a methods section for classes even if"
    " there is already an existing docstring."
    " If set it will force on entry in the section per method found."
    " If only some methods are desired to be specified then this should be left off.",
)
@click.option(
    "--force-attributes/--unforce-attributes",
    type=bool,
    is_flag=True,
    default=False,
    help="Whether to force an attributes section for classes even if"
    " there is already an existing docstring."
    " If set it will force on entry in the section per attribute found."
    " If only some attributes are desired then this should be left off.",
)
@click.option(
    "--ignore-privates/--handle-privates",
    is_flag=True,
    default=True,
    help="Whether to ignore attributes and methods that start with an underscore '_'."
    " This also means that methods with two underscores are ignored."
    " Consequently turning this off also forces processing of such methods."
    " Dunder methods are an exception and are"
    " always ignored regardless of this setting.",
)
@click.option(
    "--ignore-unused-arguments/--handle-unused-arguments",
    is_flag=True,
    default=True,
    help="Whether to ignore arguments starting with an underscore '_'"
    " are ignored when building parameter sections.",
)
@click.option(
    "--ignored-decorators",
    multiple=True,
    default=["overload"],
    help="Decorators that, if present,"
    " should cause a function to be ignored for docstring analysis and generation.",
)
@click.option(
    "--ignored-functions",
    multiple=True,
    default=["main"],
    help="Functions that should be ignored for docstring analysis and generation."
    " Only exact matches are ignored. This is not a regex pattern.",
)
@click.option(
    "--ignored-classes",
    multiple=True,
    default=[],
    help="Classes that should be ignored for docstring analysis and generation."
    " Only exact matches are ignored. This is not a regex pattern.",
)
@click.option(
    "--force-defaults/--unforce-defaults",
    is_flag=True,
    default=True,
    help="Whether to enforce descriptions need to"
    " name/explain the default value of their parameter.",
)
@click.option(
    "-q",
    "--quiet",
    is_flag=True,
    help=(
        "Don't emit non-error messages to stderr. Errors are still emitted; silence"
        " those with 2>/dev/null."
    ),
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help=(
        "Also emit messages to stderr about files that were not changed or were ignored"
        " due to exclusion patterns."
    ),
)
@click.version_option(
    version=__version__,
    message=(
        f"%(prog)s, %(version)s\n"
        f"Python ({platform.python_implementation()}) {platform.python_version()}"
    ),
)
@click.argument(
    "src",
    nargs=-1,
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, allow_dash=False
    ),
    is_eager=True,
    metavar="SRC ...",
)
@click.option(
    "--config",
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        allow_dash=False,
        path_type=str,
    ),
    is_eager=True,
    callback=read_pyproject_toml,
    help="Read configuration from FILE path.",
)
@click.pass_context
def main(  # pylint: disable=too-many-arguments, too-many-locals  # noqa: PLR0913
    ctx: click.Context,
    *,
    write: bool,
    output_style: dsp.DocstringStyle,
    input_style: dsp.DocstringStyle,
    check: bool,
    exclude: Optional[Pattern[str]],
    extend_exclude: Optional[Pattern[str]],
    force_params: bool,
    force_params_min_n_params: bool,
    force_meta_min_func_length: bool,
    force_return: bool,
    force_raises: bool,
    force_methods: bool,
    force_attributes: bool,
    ignore_privates: bool,
    ignore_unused_arguments: bool,
    ignored_decorators: list[str],
    ignored_functions: list[str],
    ignored_classes: list[str],
    force_defaults: bool,
    quiet: bool,
    verbose: bool,
    src: tuple[str, ...],
    config: Optional[str],
) -> None:
    """Create, update or convert docstrings."""
    ctx.ensure_object(dict)

    if not src:
        out(main.get_usage(ctx) + "\n\nError: Missing argument 'SRC ...'.")
        ctx.exit(1)

    if verbose and config:
        config_source = ctx.get_parameter_source("config")
        if config_source in (
            ParameterSource.DEFAULT,
            ParameterSource.DEFAULT_MAP,
        ):
            out("Using configuration from project root.", fg="blue")
        else:
            out(f"Using configuration in '{config}'.", fg="blue")
        if ctx.default_map:
            for param, value in ctx.default_map.items():
                out(f"{param}: {value}")

    report = Report(check=check, diff=not write, quiet=quiet, verbose=verbose)
    fixer_settings = FixerSettings(
        force_params=force_params,
        force_return=force_return,
        force_raises=force_raises,
        force_methods=force_methods,
        force_attributes=force_attributes,
        force_params_min_n_params=force_params_min_n_params,
        force_meta_min_func_length=force_meta_min_func_length,
        ignore_privates=ignore_privates,
        ignore_unused_arguments=ignore_unused_arguments,
        ignored_decorators=ignored_decorators,
        ignored_functions=ignored_functions,
        ignored_classes=ignored_classes,
        force_defaults=force_defaults,
    )

    run(
        src,
        overwrite=write,
        output_style=output_style,
        input_style=input_style,
        exclude=exclude or DEFAULT_EXCLUDES,
        extend_exclude=extend_exclude,
        report=report,
        fixer_settings=fixer_settings,
    )

    if verbose or not quiet:
        if verbose or report.change_count or report.failure_count:
            out()
        error_msg = "Oh no! 💥 💔 💥"
        out(error_msg if report.return_code else "All done! ✨ 🍰 ✨")
        click.echo(str(report), err=True)
    ctx.exit(report.return_code)



"""Epyoc-style docstring parsing.

.. seealso:: http://epydoc.sourceforge.net/manual-fields.html
"""
import inspect
import re
from typing import NamedTuple, Optional

from .common import (
    Docstring,
    DocstringMeta,
    DocstringParam,
    DocstringRaises,
    DocstringReturns,
    DocstringStyle,
    DocstringYields,
    ParseError,
    RenderingStyle,
    append_description,
    clean_str,
    split_description,
)


class SectionPattern(NamedTuple):
    """Patterns for docstring sections."""

    param: re.Pattern[str]
    raises: re.Pattern[str]
    returns: re.Pattern[str]
    meta: re.Pattern[str]


class SectionMatch(NamedTuple):
    """Matches of docstring sections."""

    param: Optional[re.Match[str]]
    raises: Optional[re.Match[str]]
    returns: Optional[re.Match[str]]
    meta: Optional[re.Match[str]]


def _get_matches_for_chunk(chunk: str, patterns: SectionPattern) -> SectionMatch:
    """Apply a search for each pattern to the chunk.

    Parameters
    ----------
    chunk : str
        Chunk to match the patterns against.
    patterns : SectionPattern
        Collection of regex patterns to match against the chunk.

    Returns
    -------
    SectionMatch
        Tuple of matches of the patterns against the chunk.
    """
    return SectionMatch(
        param=re.search(patterns.param, chunk),
        raises=re.search(patterns.raises, chunk),
        returns=re.search(patterns.returns, chunk),
        meta=re.search(patterns.meta, chunk),
    )


class StreamToken(NamedTuple):
    """One entry of the stream list."""

    base: str
    key: str
    args: list[str]
    desc: str


def _tokenize(
    meta_chunk: str,
    patterns: SectionPattern,
) -> list[StreamToken]:
    """Return the tokenized stream according to the regex patterns.

    Parameters
    ----------
    meta_chunk : str
        Chunk to tokenize.
    patterns : SectionPattern
        Collection of patterns for different sections.

    Returns
    -------
    list[StreamToken]
        (base, key, args, desc)
        base: Literal['param', 'raise', 'return', 'meta']
        key: str:
        args: List[str]
        desc: str: Description

    Raises
    ------
    ParseError
        If none of the patterns match against the chunk.
    ParseError
        If we match a section in the general meta case that should have already
        been matched in a specific section.
    """
    stream: list[StreamToken] = []
    for chunk_match in re.finditer(r"(^@.*?)(?=^@|\Z)", meta_chunk, flags=re.S | re.M):
        chunk = chunk_match.group(0)
        if not chunk:
            continue

        matches = _get_matches_for_chunk(chunk, patterns)

        match = matches.param or matches.raises or matches.returns or matches.meta
        if not match:
            msg = f'Error parsing meta information near "{chunk}".'
            raise ParseError(msg)

        if matches.param:
            base = "param"
            key: str = match.group(1)
            args = [match.group(2).strip()]
        elif matches.raises:
            base = "raise"
            key: str = match.group(1)
            args = [] if match.group(2) is None else [match.group(2).strip()]
        elif matches.returns:
            base = "return" if match.group(1) in ("return", "rtype") else "yield"
            key: str = match.group(1)
            args = []
        else:
            base = "meta"
            key: str = match.group(1)
            token = clean_str(match.group(2).strip())
            args = [] if token is None else re.split(r"\s+", token)

            # Make sure we didn't match some existing keyword in an incorrect
            # way here:
            if key in {
                "param",
                "keyword",
                "type",
                "return",
                "rtype",
                "yield",
                "ytype",
            }:
                msg = f'Error parsing meta information near "{chunk}".'
                raise ParseError(msg)

        desc = chunk[match.end() :].strip()
        if "\n" in desc:
            first_line, rest = desc.split("\n", 1)
            desc = first_line + "\n" + inspect.cleandoc(rest)
        stream.append(StreamToken(base, key, args, desc))
    return stream


def _combine_params(stream: list[StreamToken]) -> dict[str, dict[str, Optional[str]]]:
    """Group the list of tokens into sections based on section and information..

    Parameters
    ----------
    stream : list[StreamToken]
        List of tokens to group into dict.

    Returns
    -------
    dict[str, dict[str, Optional[str]]]
        Dictionary grouping parsed param sections
        by section (param name, "return", "yield") and
        information they represent (type_name, description)
    """
    params: dict[str, dict[str, Optional[str]]] = {}
    for base, key, args, desc in stream:
        if base not in ["param", "return", "yield"]:
            continue  # nothing to do
        arg_name = args[0] if base == "param" else base
        info = params.setdefault(arg_name, {})
        info_key = "type_name" if "type" in key else "description"
        info[info_key] = desc
    return params


def _add_meta_information(
    stream: list[StreamToken],
    params: dict[str, dict[str, Optional[str]]],
    ret: Docstring,
) -> None:
    """Add the meta information into the docstring instance.

    Parameters
    ----------
    stream : list[StreamToken]
        Stream of tokens of the string-
    params : dict[str, dict[str, Optional[str]]]
        Grouped information about each section.
    ret : Docstring
        Docstring instance to add the information to.

    Raises
    ------
    ParseError
        If an unexpected section is encountered.
    """
    is_done: dict[str, bool] = {}
    for token in stream:
        if token.base == "param" and not is_done.get(token.args[0], False):
            (arg_name,) = token.args
            info = params[arg_name]
            type_name = info.get("type_name")

            if type_name and type_name.endswith("?"):
                is_optional = True
                type_name = type_name[:-1]
            else:
                is_optional = False

            match = re.match(r".*defaults to (.+)", token.desc, flags=re.DOTALL)
            default = match[1].rstrip(".") if match else None

            meta_item = DocstringParam(
                args=[token.key, arg_name],
                description=info.get("description"),
                arg_name=arg_name,
                type_name=type_name,
                is_optional=is_optional,
                default=default,
            )
            is_done[arg_name] = True
        elif token.base == "return" and not is_done.get("return", False):
            info = params["return"]
            meta_item = DocstringReturns(
                args=[token.key],
                description=info.get("description"),
                type_name=info.get("type_name"),
                is_generator=False,
            )
            is_done["return"] = True
        elif token.base == "yield" and not is_done.get("yield", False):
            info = params["yield"]
            meta_item = DocstringYields(
                args=[token.key],
                description=info.get("description"),
                type_name=info.get("type_name"),
                is_generator=True,
            )
            is_done["yield"] = True
        elif token.base == "raise":
            (type_name,) = token.args or (None,)
            meta_item = DocstringRaises(
                args=[token.key, *token.args],
                description=token.desc,
                type_name=type_name,
            )
        elif token.base == "meta":
            meta_item = DocstringMeta(
                args=[token.key, *token.args],
                description=token.desc,
            )
        else:
            arg_key = token.args[0] if token.args else token.base
            if not is_done.get(arg_key, False):
                msg = (
                    "Error building meta information. "
                    f"Encountered unexpected section {arg_key}."
                )
                raise ParseError(msg)
            continue  # don't append

        ret.meta.append(meta_item)


def parse(text: Optional[str]) -> Docstring:
    """Parse the epydoc-style docstring into its components.

    Parameters
    ----------
    text : Optional[str]
        docstring to parse

    Returns
    -------
    Docstring
        parsed docstring
    """
    ret = Docstring(style=DocstringStyle.EPYDOC)
    if not text:
        return ret

    text = inspect.cleandoc(text)
    if match := re.search("^@", text, flags=re.M):
        desc_chunk = text[: match.start()]
        meta_chunk = text[match.start() :]
    else:
        desc_chunk = text
        meta_chunk = ""

    split_description(ret, desc_chunk)

    patterns = SectionPattern(
        param=re.compile(r"(param|keyword|type)(\s+[_A-z][_A-z0-9]*\??):"),
        raises=re.compile(r"(raise)(\s+[_A-z][_A-z0-9]*\??)?:"),
        returns=re.compile(r"(return|rtype|yield|ytype):"),
        meta=re.compile(r"([_A-z][_A-z0-9]+)((\s+[_A-z][_A-z0-9]*\??)*):"),
    )

    # tokenize
    stream = _tokenize(meta_chunk, patterns)

    # Combine type_name, arg_name, and description information
    params = _combine_params(stream)

    _add_meta_information(stream, params, ret)

    return ret


def compose(
    docstring: Docstring,
    rendering_style: RenderingStyle = RenderingStyle.COMPACT,
    indent: str = "    ",
) -> str:
    """Render a parsed docstring into docstring text.

    Parameters
    ----------
    docstring : Docstring
        parsed docstring representation
    rendering_style : RenderingStyle
        the style to render docstrings (Default value = RenderingStyle.COMPACT)
    indent : str
        the characters used as indentation in the
        docstring string (Default value = '    ')

    Returns
    -------
    str
        docstring text
    """

    def process_desc(desc: Optional[str], *, is_type: bool) -> str:
        """Process a description section.

        Parameters
        ----------
        desc : Optional[str]
            Description to process
        is_type : bool
            Whether the description represent type information.

        Returns
        -------
        str
            The properly rendered description information.
        """
        if not desc:
            return ""

        if rendering_style == RenderingStyle.EXPANDED or (
            rendering_style == RenderingStyle.CLEAN and not is_type
        ):
            (first, *rest) = desc.splitlines()
            return "\n".join(["\n" + indent + first] + [indent + line for line in rest])

        (first, *rest) = desc.splitlines()
        return "\n".join([f" {first}"] + [indent + line for line in rest])

    parts: list[str] = []
    append_description(docstring, parts)

    for meta in docstring.meta:
        if isinstance(meta, DocstringParam):
            if meta.type_name:
                type_name = f"{meta.type_name}?" if meta.is_optional else meta.type_name
                text = f"@type {meta.arg_name}:"
                text += process_desc(type_name, is_type=True)
                parts.append(text)
            text = (
                f"@param {meta.arg_name}:"
                f"{process_desc(meta.description, is_type=False)}"
            )
            parts.append(text)
        elif isinstance(meta, (DocstringReturns, DocstringYields)):
            (arg_key, type_key) = (
                ("yield", "ytype")
                if isinstance(meta, DocstringYields)
                else ("return", "rtype")
            )
            if meta.type_name:
                text = f"@{type_key}:{process_desc(meta.type_name, is_type=True)}"
                parts.append(text)
            if meta.description:
                text = f"@{arg_key}:{process_desc(meta.description, is_type=False)}"
                parts.append(text)
        elif isinstance(meta, DocstringRaises):
            text = f"@raise {meta.type_name}:" if meta.type_name else "@raise:"
            text += process_desc(meta.description, is_type=False)
            parts.append(text)
        else:
            text = f'@{" ".join(meta.args)}:'
            text += process_desc(meta.description, is_type=False)
            parts.append(text)
    return "\n".join(parts)



"""ReST-style docstring parsing."""

import inspect
import re
from typing import Optional, Union

from .common import (
    DEPRECATION_KEYWORDS,
    PARAM_KEYWORDS,
    RAISES_KEYWORDS,
    RETURNS_KEYWORDS,
    YIELDS_KEYWORDS,
    Docstring,
    DocstringDeprecated,
    DocstringMeta,
    DocstringParam,
    DocstringRaises,
    DocstringReturns,
    DocstringStyle,
    DocstringYields,
    ParseError,
    RenderingStyle,
    append_description,
    split_description,
)


def _build_param(args: list[str], desc: str) -> DocstringParam:
    """Build parameter entry from supplied arguments.

    Parameters
    ----------
    args : list[str]
        List of strings describing the parameter (name, type)
    desc : str
        String representing the parameter description.

    Returns
    -------
    DocstringParam
        The docstring object combining and structuring the raw info.

    Raises
    ------
    ParseError
        If an unexpected number of arguments were found.
    """
    if len(args) == 3:
        _, type_name, arg_name = args
        if type_name.endswith("?"):
            is_optional = True
            type_name = type_name[:-1]
        else:
            is_optional = False
    elif len(args) == 2:
        _, arg_name = args
        type_name = None
        is_optional = None
    else:
        msg = f"Expected one or two arguments for a {args[0]} keyword."
        raise ParseError(msg)

    match = re.match(r".*defaults to (.+)", desc, flags=re.DOTALL)
    default = match[1].rstrip(".") if match else None

    return DocstringParam(
        args=args,
        description=desc,
        arg_name=arg_name,
        type_name=type_name,
        is_optional=is_optional,
        default=default,
    )


def _build_return(args: list[str], desc: str) -> DocstringReturns:
    """Build return entry from supplied arguments.

    Parameters
    ----------
    args : list[str]
        List of strings describing the return value (name, type)
    desc : str
        String representing the return description.

    Returns
    -------
    DocstringReturns
        The docstring object combining and structuring the raw info.

    Raises
    ------
    ParseError
        If an unexpected number of arguments were found.
    """
    if len(args) == 2:
        type_name = args[1]
    elif len(args) == 1:
        type_name = None
    else:
        msg = f"Expected one or no arguments for a {args[0]} keyword."
        raise ParseError(msg)

    return DocstringReturns(
        args=args,
        description=desc,
        type_name=type_name,
        is_generator=False,
    )


def _build_yield(args: list[str], desc: str) -> DocstringYields:
    """Build yield entry from supplied arguments.

    Parameters
    ----------
    args : list[str]
        List of strings describing the yielded value (name, type)
    desc : str
        String representing the yield value description.

    Returns
    -------
    DocstringYields
        The docstring object combining and structuring the raw info.

    Raises
    ------
    ParseError
        If an unexpected number of arguments were found.
    """
    if len(args) == 2:
        type_name = args[1]
    elif len(args) == 1:
        type_name = None
    else:
        msg = f"Expected one or no arguments for a {args[0]} keyword."
        raise ParseError(msg)

    return DocstringYields(
        args=args,
        description=desc,
        type_name=type_name,
        is_generator=True,
    )


def _build_deprecation(args: list[str], desc: str) -> DocstringDeprecated:
    """Build deprecation entry from supplied arguments.

    Parameters
    ----------
    args : list[str]
        List of strings describing the deprecation
    desc : str
        Actual textual description.

    Returns
    -------
    DocstringDeprecated
        The docstring object combining and structuring the raw info.
    """
    match = re.search(
        r"^(?P<version>v?((?:\d+)(?:\.[0-9a-z\.]+))) (?P<desc>.+)",
        desc,
        flags=re.I,
    )
    return DocstringDeprecated(
        args=args,
        version=match["version"] if match else None,
        description=match["desc"] if match else desc,
    )


def _build_raises(args: list[str], desc: str) -> DocstringRaises:
    """Build raises entry from supplied arguments.

    Parameters
    ----------
    args : list[str]
        List of strings describing the raised value (name, type)
    desc : str
        String representing the raised value description.

    Returns
    -------
    DocstringRaises
        The docstring object combining and structuring the raw info.

    Raises
    ------
    ParseError
        If an unexpected number of arguments were found.
    """
    if len(args) == 2:
        type_name = args[1]
    elif len(args) == 1:
        type_name = None
    else:
        msg = f"Expected one or no arguments for a {args[0]} keyword."
        raise ParseError(msg)
    return DocstringRaises(args=args, description=desc, type_name=type_name)


def _build_meta(args: list[str], desc: str) -> DocstringMeta:
    """Build a fottomg meta entry from supplied arguments.

    Parameters
    ----------
    args : list[str]
        List of strings describing entry.
    desc : str
        String representing the entry description.

    Returns
    -------
    DocstringMeta
        The docstring object combining and structuring the raw info.
    """
    key = args[0]

    if key in PARAM_KEYWORDS:
        return _build_param(args, desc)

    if key in RETURNS_KEYWORDS:
        return _build_return(args, desc)

    if key in YIELDS_KEYWORDS:
        return _build_yield(args, desc)

    if key in DEPRECATION_KEYWORDS:
        return _build_deprecation(args, desc)

    if key in RAISES_KEYWORDS:
        return _build_raises(args, desc)

    return DocstringMeta(args=args, description=desc)


def _get_chunks(text: str) -> tuple[str, str]:
    """Split the text into args (key, type, ...) and description.

    Parameters
    ----------
    text : str
        Text to split into chunks.

    Returns
    -------
    tuple[str, str]
        Args and description.
    """
    if match := re.search("^:", text, flags=re.M):
        return text[: match.start()], text[match.start() :]
    return text, ""


def _get_split_chunks(chunk: str) -> tuple[list[str], str]:
    """Split a entry into args and description.

    Parameters
    ----------
    chunk : str
        Entry string to split.

    Returns
    -------
    tuple[list[str], str]
        Arguments of the entry and its description.

    Raises
    ------
    ParseError
        If the chunk could not be split into args and description.
    """
    try:
        args_chunk, desc_chunk = chunk.lstrip(":").split(":", 1)
    except ValueError as ex:
        msg = f'Error parsing meta information near "{chunk}".'
        raise ParseError(msg) from ex
    return args_chunk.split(), desc_chunk.strip()


def _extract_type_info(
    docstring: Docstring, meta_chunk: str
) -> tuple[dict[str, str], dict[Optional[str], str], dict[Optional[str], str]]:
    """Extract type and description pairs and add other entries directly.

    Parameters
    ----------
    docstring : Docstring
        Docstring wrapper to add information to.
    meta_chunk : str
        Docstring text to extract information from.

    Returns
    -------
    types : dict[str, str]
        Dictionary matching parameters to their descriptions
    rtypes : dict[Optional[str], str]
        Dictionary matching return values to their descriptions
    ytypes : dict[Optional[str], str]
        Dictionary matching yielded values to their descriptions
    """
    types: dict[str, str] = {}
    rtypes: dict[Optional[str], str] = {}
    ytypes: dict[Optional[str], str] = {}
    for chunk_match in re.finditer(r"(^:.*?)(?=^:|\Z)", meta_chunk, flags=re.S | re.M):
        chunk = chunk_match.group(0)
        if not chunk:
            continue

        args, desc = _get_split_chunks(chunk)

        if "\n" in desc:
            first_line, rest = desc.split("\n", 1)
            desc = first_line + "\n" + inspect.cleandoc(rest)

        # Add special handling for :type a: typename
        if len(args) == 2 and args[0] == "type":
            types[args[1]] = desc
        elif len(args) in {1, 2} and args[0] == "rtype":
            rtypes[None if len(args) == 1 else args[1]] = desc
        elif len(args) in {1, 2} and args[0] == "ytype":
            ytypes[None if len(args) == 1 else args[1]] = desc
        else:
            docstring.meta.append(_build_meta(args, desc))
    return types, rtypes, ytypes


def parse(text: Optional[str]) -> Docstring:
    """Parse the ReST-style docstring into its components.

    Parameters
    ----------
    text : Optional[str]
        docstring to parse

    Returns
    -------
    Docstring
        parsed docstring

    Raises
    ------
    ParseError
        If a section does not have two colons to be split on.
    """
    ret = Docstring(style=DocstringStyle.REST)
    if not text:
        return ret

    text = inspect.cleandoc(text)
    desc_chunk, meta_chunk = _get_chunks(text)

    split_description(ret, desc_chunk)

    types, rtypes, ytypes = _extract_type_info(ret, meta_chunk)

    for meta in ret.meta:
        if isinstance(meta, DocstringParam):
            meta.type_name = meta.type_name or types.get(meta.arg_name)
        elif isinstance(meta, DocstringReturns):
            meta.type_name = meta.type_name or rtypes.get(meta.return_name)
        elif isinstance(meta, DocstringYields):
            meta.type_name = meta.type_name or ytypes.get(meta.yield_name)

    if not any(isinstance(m, DocstringReturns) for m in ret.meta) and rtypes:
        for return_name, type_name in rtypes.items():
            ret.meta.append(
                DocstringReturns(
                    args=[],
                    type_name=type_name,
                    description=None,
                    is_generator=False,
                    return_name=return_name,
                )
            )

    return ret


def process_desc(
    desc: Optional[str], rendering_style: RenderingStyle, indent: str = "    "
) -> str:
    """Process the description for one element.

    Parameters
    ----------
    desc : Optional[str]
        Description to process
    rendering_style : RenderingStyle
        Rendering style to use.
    indent : str
        Indentation needed for that line (Default value = '    ')

    Returns
    -------
    str
        String representation of the docstrings description.
    """
    if not desc:
        return ""

    if rendering_style == RenderingStyle.CLEAN:
        (first, *rest) = desc.splitlines()
        return "\n".join([f" {first}"] + [indent + line for line in rest])

    if rendering_style == RenderingStyle.EXPANDED:
        (first, *rest) = desc.splitlines()
        return "\n".join(["\n" + indent + first] + [indent + line for line in rest])

    return f" {desc}"


def _append_param(
    param: DocstringParam,
    parts: list[str],
    rendering_style: RenderingStyle,
    indent: str,
) -> None:
    """Append one parameter entry to the output stream.

    Parameters
    ----------
    param : DocstringParam
        Structured representation of a parameter entry.
    parts : list[str]
        List of strings representing the final output of compose().
    rendering_style : RenderingStyle
        Rendering style to use.
    indent : str
        Indentation needed for that line.
    """
    if param.type_name:
        type_text = (
            f" {param.type_name}? " if param.is_optional else f" {param.type_name} "
        )
    else:
        type_text = " "
    if rendering_style == RenderingStyle.EXPANDED:
        text = f":param {param.arg_name}:"
        text += process_desc(param.description, rendering_style, indent)
        parts.append(text)
        if type_text[:-1]:
            parts.append(f":type {param.arg_name}:{type_text[:-1]}")
    else:
        text = f":param{type_text}{param.arg_name}:"
        text += process_desc(param.description, rendering_style, indent)
        parts.append(text)


def _append_return(
    meta: Union[DocstringReturns, DocstringYields],
    parts: list[str],
    rendering_style: RenderingStyle,
    indent: str,
) -> None:
    """Append one return/yield entry to the output stream.

    Parameters
    ----------
    meta : Union[DocstringReturns, DocstringYields]
        Structured representation of a return/yield entry.
    parts : list[str]
        List of strings representing the final output of compose().
    rendering_style : RenderingStyle
        Rendering style to use.
    indent : str
        Indentation needed for that line.
    """
    type_text = f" {meta.type_name}" if meta.type_name else ""
    key = "yields" if isinstance(meta, DocstringYields) else "returns"

    if rendering_style == RenderingStyle.EXPANDED:
        if meta.description:
            text = f":{key}:"
            text += process_desc(meta.description, rendering_style, indent)
            parts.append(text)
        if type_text:
            return_key = "rtype" if isinstance(meta, DocstringReturns) else "ytype"
            parts.append(f":{return_key}:{type_text}")
    else:
        text = f":{key}{type_text}:"
        text += process_desc(meta.description, rendering_style, indent)
        parts.append(text)


def compose(
    docstring: Docstring,
    rendering_style: RenderingStyle = RenderingStyle.COMPACT,
    indent: str = "    ",
) -> str:
    """Render a parsed docstring into docstring text.

    Parameters
    ----------
    docstring : Docstring
        parsed docstring representation
    rendering_style : RenderingStyle
        the style to render docstrings (Default value = RenderingStyle.COMPACT)
    indent : str
        the characters used as indentation in the docstring string
        (Default value = '    ')

    Returns
    -------
    str
        docstring text
    """
    parts: list[str] = []
    append_description(docstring, parts)

    for meta in docstring.meta:
        if isinstance(meta, DocstringParam):
            _append_param(meta, parts, rendering_style, indent)
        elif isinstance(meta, (DocstringReturns, DocstringYields)):
            _append_return(meta, parts, rendering_style, indent)
        elif isinstance(meta, DocstringRaises):
            type_text = f" {meta.type_name} " if meta.type_name else ""
            text = (
                f":raises{type_text}:"
                f"{process_desc(meta.description, rendering_style, indent)}"
            )
            parts.append(text)
        else:
            text = (
                f":{' '.join(meta.args)}:"
                f"{process_desc(meta.description, rendering_style, indent)}"
            )
            parts.append(text)
    return "\n".join(parts)



"""Nice output for pymend.

The double calls are for patching purposes in tests.
"""

import difflib
import tempfile
from typing import Any, Optional

from click import echo, style


def out(
    message: Optional[str] = None, *, nl: bool = True, **styles: Any  # noqa: ANN401
) -> None:
    """Output normal message.

    Parameters
    ----------
    message : Optional[str]
        Message to output (Default value = None)
    nl : bool
        Whether to print a newline after the message. (Default value = True)
    **styles : Any
        Style options for click.
    """
    if message is not None:
        if "bold" not in styles:
            styles["bold"] = True
        message = style(message, **styles)
    echo(message, nl=nl, err=True)


def err(
    message: Optional[str] = None, *, nl: bool = True, **styles: Any  # noqa: ANN401
) -> None:
    """Output error message.

    Parameters
    ----------
    message : Optional[str]
        Message to output (Default value = None)
    nl : bool
        Whether to print a newline after the message. (Default value = True)
    **styles : Any
        Style options for click.
    """
    if message is not None:
        if "fg" not in styles:
            styles["fg"] = "red"
        message = style(message, **styles)
    echo(message, nl=nl, err=True)


def diff(a: list[str], b: list[str], a_name: str, b_name: str) -> list[str]:
    """Return a unified diff list between lists`a` and `b`.

    Parameters
    ----------
    a : list[str]
        Source for the diff
    b : list[str]
        Target for the diff.
    a_name : str
        Path to the source file of the diff. (Default value = '')
    b_name : str
        Path to the target file of the diff. (Default value = '')

    Returns
    -------
    list[str]
        The resulting diff
    """
    diff_lines: list[str] = []
    for line in difflib.unified_diff(a, b, fromfile=a_name, tofile=b_name):
        # Work around https://bugs.python.org/issue2142
        # See:
        # https://www.gnu.org/software/diffutils/manual/html_node/Incomplete-Lines.html
        if line[-1] == "\n":
            diff_lines.append(line)
        else:
            diff_lines.append(line + "\n")
            diff_lines.append("\\ No newline at end of file\n")
    return diff_lines


def color_diff(contents: str) -> str:
    """Inject the ANSI color codes to the diff.

    Parameters
    ----------
    contents : str
        Diff content to color.

    Returns
    -------
    str
        Colored diff.
    """
    lines = contents.split("\n")
    for i, line in enumerate(lines):
        if line.startswith(("+++", "---")):
            line = "\033[1m" + line + "\033[0m"  # bold, reset  # noqa: PLW2901
        elif line.startswith("@@"):
            line = "\033[36m" + line + "\033[0m"  # cyan, reset  # noqa: PLW2901
        elif line.startswith("+"):
            line = "\033[32m" + line + "\033[0m"  # green, reset # noqa: PLW2901
        elif line.startswith("-"):
            line = "\033[31m" + line + "\033[0m"  # red, reset # noqa: PLW2901
        lines[i] = line
    return "\n".join(lines)


def dump_to_file(*output: str, ensure_final_newline: bool = True) -> str:
    """Dump `output` to a temporary file. Return path to the file.

    Parameters
    ----------
    *output : str
        List of strings to dump into the output.
    ensure_final_newline : bool
        Whether to make sure that every dumped string
        ends in a new line. (Default value = True)

    Returns
    -------
    str
        Path to the produced temp file.
    """
    with tempfile.NamedTemporaryFile(
        mode="w", prefix="blk_", suffix=".log", delete=False, encoding="utf8"
    ) as f:
        for lines in output:
            f.write(lines)
            if ensure_final_newline and lines and lines[-1] != "\n":
                f.write("\n")
    return f.name



"""_summary_."""
def generator() -> Tuple[int, int, str]:
    """_summary_.

    Returns
    -------
    a : int
        Something
    b : int
        Something else
    
    Yields
    ------
    x : int
        _description_
    z : str
        desc.
    """
    if False:
        return a, b, c
    yield x, y, z


def generator() -> str:
    """_summary_.

    Returns
    -------
    _type_
        desc
    """
    if False:
        return a
    yield b

def generator() -> str:
    """_summary_.

    Returns
    -------
    _type_
        desc
    """
    return a

def generator() -> Iterable[str]:
    """_summary_."""
    yield "test"

def generator() -> Iterator[str]:
    """_summary_.

    Yields
    ------
    Nope
        _description_
    """
    pass

def generator() -> Generator[int, float, str]:
    """_summary_."""
    if False:
        return a
    yield b



"""Attribute docstrings parsing.

.. seealso:: https://peps.python.org/pep-0257/#what-is-a-docstring
"""

import ast
import inspect
import textwrap
from types import ModuleType
from typing import Any, Optional, Union, overload

from typing_extensions import TypeGuard, override

from .common import Docstring, DocstringParam

ast_constant_attr = {
    ast.Constant: "value",
    # python <= 3.7:
    ast.NameConstant: "value",
    ast.Num: "n",
    ast.Str: "s",
}


def ast_get_constant_value(
    node: Union[ast.Str, ast.Num, ast.NameConstant, ast.Constant]
) -> Any:  # noqa: ANN401
    """Return the constant's value if the given node is a constant.

    Parameters
    ----------
    node : Union[ast.Str, ast.Num, ast.NameConstant, ast.Constant]
        Node to get the constant value from.

    Returns
    -------
    Any
        Actual value contained by the node.
    """
    return getattr(node, ast_constant_attr[node.__class__])


@overload
def ast_unparse(node: None) -> None:
    ...


@overload
def ast_unparse(node: ast.AST) -> str:
    ...


def ast_unparse(node: Optional[ast.AST]) -> Optional[str]:
    """Convert the AST node to source code as a string.

    Parameters
    ----------
    node : Optional[ast.AST]
        Node to unparse.

    Returns
    -------
    Optional[str]
        `None` if `node` was `None`.
        Otherwise the unparsed node.
    """
    if node is None:
        return None
    return ast.unparse(node)


def ast_is_literal_str(node: ast.AST) -> TypeGuard[ast.Expr]:
    """Return True if the given node is a literal string.

    Parameters
    ----------
    node : ast.AST
        Node to check.

    Returns
    -------
    TypeGuard[ast.Expr]
        True if the node represents a string literal.
    """
    return (
        isinstance(node, ast.Expr)
        and isinstance(node.value, (ast.Constant, ast.Str))
        and isinstance(ast_get_constant_value(node.value), str)
    )


def ast_get_attribute(
    node: ast.AST,
) -> Optional[tuple[str, Optional[str], Optional[str]]]:
    """Return name, type and default if the given node is an attribute.

    Parameters
    ----------
    node : ast.AST
        Node to get the attribute from

    Returns
    -------
    Optional[tuple[str, Optional[str], Optional[str]]]
        None if the node does not represent an assignment.
        Otherwise a tuple of name, type and default value.
    """
    if isinstance(node, (ast.Assign, ast.AnnAssign)):
        target = node.targets[0] if isinstance(node, ast.Assign) else node.target
        if isinstance(target, ast.Name):
            type_str = None
            if isinstance(node, ast.AnnAssign):
                type_str = ast_unparse(node.annotation)
            default = ast_unparse(node.value) if node.value else None
            return target.id, type_str, default
    return None


class AttributeDocstrings(ast.NodeVisitor):
    """An ast.NodeVisitor that collects attribute docstrings."""

    attr_docs: Optional[dict[str, tuple[str, Optional[str], Optional[str]]]] = None
    prev_attr = None

    @override
    def visit(self, node: ast.AST) -> None:
        """Visit a node and collect its attribute docstrings.

        Parameters
        ----------
        node : ast.AST
            Node to visit.
        """
        if self.prev_attr and self.attr_docs is not None and ast_is_literal_str(node):
            attr_name, attr_type, attr_default = self.prev_attr
            # This is save because `ast_is_literal_str`
            # ensure that node.value is of type (ast.Constant, ast.Str)
            self.attr_docs[attr_name] = (
                ast_get_constant_value(
                    node.value  # pyright: ignore[reportGeneralTypeIssues]
                ),
                attr_type,
                attr_default,
            )
        self.prev_attr = ast_get_attribute(node)
        if isinstance(node, (ast.ClassDef, ast.Module)):
            self.generic_visit(node)

    def get_attr_docs(
        self, component: Any  # noqa: ANN401
    ) -> dict[str, tuple[str, Optional[str], Optional[str]]]:
        """Get attribute docstrings from the given component.

        Parameters
        ----------
        component : Any
            component to process (class or module)

        Returns
        -------
        dict[str, tuple[str, Optional[str], Optional[str]]]
            for each attribute docstring, a tuple with
            (description, type, default)
        """
        self.attr_docs = {}
        self.prev_attr = None
        try:
            source = textwrap.dedent(inspect.getsource(component))
        except OSError:
            pass
        else:
            # This change might cause issues with older python versions
            # Not sure yet.
            tree = ast.parse(source)
            self.visit(tree)
        return self.attr_docs


def add_attribute_docstrings(
    obj: Union[type, ModuleType], docstring: Docstring
) -> None:
    """Add attribute docstrings found in the object's source code.

    Parameters
    ----------
    obj : Union[type, ModuleType]
        object from which to parse attribute docstrings
    docstring : Docstring
        Docstring object where found attributes are added
    """
    params = {p.arg_name for p in docstring.params}
    for arg_name, (description, type_name, default) in (
        AttributeDocstrings().get_attr_docs(obj).items()
    ):
        if arg_name not in params:
            param = DocstringParam(
                args=["attribute", arg_name],
                description=description,
                arg_name=arg_name,
                type_name=type_name,
                is_optional=default is not None,
                default=default,
            )
            docstring.meta.append(param)



"""Constant value used across pymend."""

import re

DEFAULT_EXCLUDES = re.compile(
    r"/(\.direnv|\.eggs|\.git|\.hg|\.ipynb_checkpoints|\.mypy_cache|\.nox|\.pytest_cache|\.ruff_cache|\.tox|\.svn|\.venv|\.vscode|__pypackages__|_build|buck-out|build|dist|venv)/"  # pylint: disable=line-too-long
)
DEFAULT_DESCRIPTION = "_description_"
DEFAULT_TYPE = "_type_"
DEFAULT_SUMMARY = "_summary_."



"""Module for general management of writing docstrings of multiple files."""

import ast
import os
import platform
import sys
import tempfile
import traceback
from dataclasses import dataclass
from pathlib import Path
from typing import NamedTuple

from typing_extensions import Self

import pymend.docstring_parser as dsp

from .file_parser import AstAnalyzer
from .output import diff
from .report import Changed
from .types import ElementDocstring, FixerSettings

__author__ = "J-E. Nitschke"
__copyright__ = "Copyright 2012-2021 A. Daouzli"
__licence__ = "GPL3"
__version__ = "1.0.7"
__maintainer__ = "J-E. Nitschke"


@dataclass
class FileContentRepresentation:
    """Container for str and list representation of file contents."""

    lst: list[str]
    lines: str


class Styles(NamedTuple):
    """Container for input and output style."""

    input_style: dsp.DocstringStyle
    output_style: dsp.DocstringStyle


class PyComment:
    """Manage several python scripts docstrings.

    It is used to parse and rewrite in a Pythonic way all the
    functions', methods' and classes' docstrings.
    The changes are then provided in a patch file.
    """

    def __init__(
        self,
        input_file: str,
        *,
        fixer_settings: FixerSettings,
        output_style: dsp.DocstringStyle = dsp.DocstringStyle.NUMPYDOC,
        input_style: dsp.DocstringStyle = dsp.DocstringStyle.AUTO,
        proceed_directly: bool = True,
    ) -> None:
        r"""Set the configuration including the source to proceed and options.

        Parameters
        ----------
        input_file : str
            path name (file or folder)
        fixer_settings : FixerSettings
            Settings for which fixes should be performed.
        output_style : dsp.DocstringStyle
            Output style to use for docstring.
            (Default value = dsp.DocstringStyle.NUMPYDOC)
        input_style : dsp.DocstringStyle
            Input docstring style.
            Auto means that the style is detected automatically. Can cause issues when
            styles are mixed in examples or descriptions."
            (Default value = dsp.DocstringStyle.AUTO)
        proceed_directly : bool
            Whether the file should be parsed directly with the call of
            the constructor. (Default value = True)
        """
        self.input_file = input_file
        self.style = Styles(input_style, output_style)
        input_lines = Path(self.input_file).read_text(encoding="utf-8")
        self._input = FileContentRepresentation(
            input_lines.splitlines(keepends=True), input_lines
        )
        self._output = FileContentRepresentation([], "")
        self.settings = fixer_settings
        self._changed = []
        self.docs_list = []
        self.fixed = False
        if proceed_directly:
            self.proceed()

    def proceed(self) -> None:
        """Parse file and generates/converts the docstrings."""
        self._parse()
        self._compute_before_after()

    def _parse(self) -> list[ElementDocstring]:
        """Parse input file's content and generates a list of its elements/docstrings.

        Returns
        -------
        list[ElementDocstring]
            List of information about module, classes and functions.
        """
        ast_parser = AstAnalyzer(self._input.lines, settings=self.settings)
        self.docs_list = sorted(
            ast_parser.parse_from_ast(), key=lambda element: element.lines
        )
        return self.docs_list

    def _compute_before_after(self) -> tuple[list[str], list[str], list[str]]:
        r"""Compute the before and after and assert equality and stability.

        Make sure that pymend is idempotent.
        Make sure that the original and final Ast's are the same (except for docstring.)

        Returns
        -------
        tuple[list[str], list[str], list[str]]
            Tuple of before, after, changed,
        """
        list_from, list_to, list_changed = self._get_changes()

        self._output.lst = list_to
        self._output.lines = "".join(list_to)
        self._changed = list_changed

        self.assert_stability(list_from, list_to)
        self.assert_equality(self._input.lines, self._output.lines)
        self.fixed = True
        return list_from, list_to, list_changed

    def _get_changes(self) -> tuple[list[str], list[str], list[str]]:
        r"""Compute the list of lines before and after the proposed docstring changes.

        Elements of the list already contain '\n' at the end.

        Returns
        -------
        list_from : list[str]
            Original file as list of lines.
        list_to : list[str]
            Modified content as list of lines.
        list_changed : list[str]
            List of names of elements that were changed.

        Raises
        ------
        ValueError
            If the endline of a docstring was parsed as None.
        """
        list_from = self._input.lst
        list_to: list[str] = []
        list_changed: list[str] = []
        last = 0
        # Loop over all found docstrings and replace the lines where they used to
        # (or ought to) be with the new docstring.
        for e in self.docs_list:
            start, end = e.lines
            if end is None:
                msg = "End of docstring is None. Not sure what to do with this yet."
                raise ValueError(msg)
            # e.line are line number starting at one.
            # We are now using them to index into a list starting at 0.
            start, end = start - 1, end - 1
            in_docstring = e.docstring
            old_line = list_from[start]
            leading_whitespace = old_line[: -len(old_line.lstrip())]
            raw_out = e.output_docstring(
                output_style=self.style.output_style,
                input_style=self.style.input_style,
                settings=self.settings,
            )
            out_docstring = self._add_quotes_indentation_modifier(
                raw_out,
                indentation=leading_whitespace,
                modifier=e.modifier,
            )
            if in_docstring != out_docstring.strip()[3 + len(e.modifier) : -3]:
                list_changed.append(e.name)
            list_to.extend(list_from[last:start])
            list_to.extend(out_docstring.splitlines(keepends=True))
            if not in_docstring:
                list_to.append(old_line)
            last = end + 1
        if last < len(list_from):
            list_to.extend(list_from[last:])
        return list_from, list_to, list_changed

    def _add_quotes_indentation_modifier(
        self,
        docstring: str,
        quotes: str = '"""',
        indentation: str = "    ",
        modifier: str = "",
    ) -> str:
        r"""Add quotes, indentation and modifiers to the docstring.

        Parameters
        ----------
        docstring : str
            The raw docstring to complete.
        quotes : str
            Quotes to use for the docstring. (Default value = '\"\"\"')
        indentation : str
            How much to indent the docstring lines (Default value = '    ')
        modifier : str
            Modifier to put before the opening triple quotes.
            Any combination of ("r", "f", "u") (Default value = '')

        Returns
        -------
        str
            The properly indented docstring, wrapped in triple quotes
            and preceded by the desired modifier.
        """
        split = f"{modifier}{quotes}{docstring}".splitlines()
        # One line docstring get the quotes on the same line
        if len(split) > 1:
            split.append(quotes)
        # Multi-line get them on the next
        else:
            split[0] += quotes
        for index, line in enumerate(split):
            if line.strip():
                split[index] = indentation + line
        return "\n".join(split) + "\n"

    def assert_stability(self, src: list[str], dst: list[str]) -> None:
        """Assert that running pymend on its own output does not change anything.

        Parameters
        ----------
        src : list[str]
            List of lines from the input file.
        dst : list[str]
            List of lines that pymend produced.

        Raises
        ------
        AssertionError
            If a second run of pymend produces a different output than the first.
        """
        # pylint: disable=protected-access
        comment = self.__copy_from_output()
        comment._parse()  # noqa: SLF001
        before, after, changed = comment._get_changes()  # noqa: SLF001
        if changed or not (dst == before and dst == after):
            log = self.dump_to_file(
                "INTERNAL ERROR: PyMend produced different "
                "docstrings on the second pass.\n"
                "Changed:\n",
                "\n".join(changed),
                "".join(diff(src, dst, "source", "first pass")),
                "".join(diff(dst, after, "first pass", "second pass")),
            )
            msg = (
                "INTERNAL ERROR:"
                " PyMend produced different docstrings on the second pass."
                " Please report a bug on"
                " https://github.com/JanEricNitschke/pymend/issues."
                f" This diff might be helpful: {log}"
            )
            raise AssertionError(msg)

    def assert_equality(self, src_lines: str, dst_lines: str) -> None:
        """Assert that running pymend does not change functional ast.

        Done by comparing the asts for the original and produced outputs
        while ignoring the docstrings themselves.

        Parameters
        ----------
        src_lines : str
            Lines from the input file.
        dst_lines : str
            Lines that pymend produced.

        Raises
        ------
        AssertionError
            If the content of the input file could not be parsed into an ast.
        AssertionError
            If the output from pymend could not be parsed into an ast.
        AssertionError
            If the output from pymend produces a different (reduced) ast
            than the input.
        """
        try:
            src_ast = ast.parse(src_lines)
        except Exception as exc:  # noqa: BLE001
            msg = f"Failed to parse source file AST: {exc}\n"
            raise AssertionError(msg) from exc
        try:
            dst_ast = ast.parse(dst_lines)
        except Exception as exc:  # noqa: BLE001
            log = self.dump_to_file(
                "INTERNAL ERROR: PyMend produced invalid code:\n",
                "".join(traceback.format_tb(exc.__traceback__)),
                dst_lines,
            )
            msg = (
                f"INTERNAL ERROR: PyMend produced invalid code: {exc}. "
                "Please report a bug on"
                " https://github.com/JanEricNitschke/pymend/issues."
                f"  This invalid output might be helpful: {log}"
            )
            raise AssertionError(msg) from None
        src_ast_list = self._stringify_ast(src_ast)
        dst_ast_list = self._stringify_ast(dst_ast)
        if src_ast_list != dst_ast_list:
            log = self.dump_to_file(
                "INTERNAL ERROR: PyMend produced code "
                "that is not equivalent to the source\n",
                "".join(diff(src_ast_list, dst_ast_list, "src", "dst")),
            )
            msg = (
                "INTERNAL ERROR: PyMend produced code that is not equivalent to the"
                " source.  Please report a bug on "
                "https://github.com/JanEricNitschke/pymend/issues."
                f"  This diff might be helpful: {log}"
            )
            raise AssertionError(msg) from None

    def __copy_from_output(self) -> Self:
        """Create a new PyComment with the same output style and lines from the input.

        Parameters
        ----------
        lines : list[str]
            List of lines that should make up the `input_lines` of the copied
            instance.

        Returns
        -------
        Self
            The new instance with the same output style and lines initialized
            by the `lines` argument.
        """
        # pylint: disable=protected-access
        py_comment = PyComment.__new__(PyComment)
        py_comment._input = FileContentRepresentation(  # noqa: SLF001
            self._output.lst.copy(), self._output.lines
        )
        py_comment.settings = self.settings
        py_comment._output = FileContentRepresentation([], "")  # noqa: SLF001
        py_comment.style = self.style
        py_comment.docs_list = []
        return py_comment

    def _strip_ast(self, ast_node: ast.AST) -> None:
        """Remove all docstrings from the ast.

        Parameters
        ----------
        ast_node : ast.AST
            Node representing the full ast.
        """
        for node in ast.walk(ast_node):
            # let's work only on functions & classes definitions
            if not isinstance(
                node, (ast.FunctionDef, ast.ClassDef, ast.AsyncFunctionDef, ast.Module)
            ):
                continue

            if not node.body:
                continue

            if not isinstance(first_element := node.body[0], ast.Expr):
                continue

            if not isinstance(docnode := first_element.value, ast.Constant):
                continue

            if not isinstance(docnode.value, str):
                continue

            node.body = node.body[1:]

    def _stringify_ast(self, node: ast.AST) -> list[str]:
        """Turn ast into string representation with all docstrings removed.

        Parameters
        ----------
        node : ast.AST
            Node to turn into a reduced string representation.

        Returns
        -------
        list[str]
            List of lines making up the reduced string representation.
        """
        self._strip_ast(node)
        return ast.dump(node, indent=1).splitlines(keepends=True)

    def dump_to_file(self, *output: str, ensure_final_newline: bool = True) -> str:
        """Dump `output` to a temporary file. Return path to the file.

        Parameters
        ----------
        *output : str
            List of strings to dump into the output.
        ensure_final_newline : bool
            Whether to make sure that every dumped string
            ends in a new line. (Default value = True)

        Returns
        -------
        str
            Path to the produced temp file.
        """
        with tempfile.NamedTemporaryFile(
            mode="w", prefix="blk_", suffix=".log", delete=False, encoding="utf8"
        ) as f:
            for lines in output:
                f.write(lines)
                if ensure_final_newline and lines and lines[-1] != "\n":
                    f.write("\n")
        return f.name

    def _docstring_diff(self) -> list[str]:
        """Build the diff between original docstring and proposed docstring.

        Returns
        -------
        list[str]
            The resulting diff
        """
        return diff(
            self._input.lst,
            self._output.lst,
            f"a/{self.input_file}",
            f"b/{self.input_file}",
        )

    def output_patch(self) -> Changed:
        """Output the patch. Either to stdout or a file depending on input file.

        Returns
        -------
        Changed
            Whether there were any changes.
        """
        if not self.fixed:
            self.proceed()
        if self._changed:
            lines_to_write = self._get_patch_lines()

            if self.input_file == "-":
                sys.stdout.writelines(lines_to_write)
            else:
                self._write_patch_file(lines_to_write)
        return Changed.YES if bool(self._changed) else Changed.NO

    def output_fix(self) -> Changed:
        """Output the fixed file. Either to stdout or the file.

        Returns
        -------
        Changed
            Whether there were any changes.

        Raises
        ------
        AssertionError
            If the input and output lines are identical but pyment reports
            some elements to have changed.
        """
        if not self.fixed:
            self.proceed()
        if (self._input.lines == self._output.lines) != (len(self._changed) == 0):
            log = self.dump_to_file(
                "INTERNAL ERROR: "
                "Pymend reported some elements to have changed "
                "but the output is identical to the input.\n",
                "List of changed elements:\n",
                "\n".join(self._changed),
                "Original file\n",
                self._input.lines,
            )
            msg = (
                "INTERNAL ERROR: "
                "Pymend reported some elements to have changed "
                "but the output is identical to the input. "
                " Please report a bug on"
                " https://github.com/JanEricNitschke/pymend/issues."
                f" This invalid output might be helpful: {log}"
            )
            raise AssertionError(msg)
        if self.input_file == "-":
            sys.stdout.writelines(self._output.lst)
        elif self._input.lines != self._output.lines:
            print(
                "Modified docstrings of element"
                f'{"s" if len(self._changed) > 1 else ""} '
                f'({", ".join(self._changed)}) in file {self.input_file}.'
            )
            self._overwrite_source_file()
        return Changed.YES if bool(self._changed) else Changed.NO

    def _get_patch_lines(self) -> list[str]:
        r"""Return the diff between source_path and target_path.

        Parameters
        ----------
        source_path : str
            name of the original file (Default value = '')
        target_path : str
            name of the final file (Default value = '')

        Returns
        -------
        list[str]
            the diff as a list of \n terminated lines
        """
        return [
            f"# Patch generated by Pymend v{__version__}\n\n",
            *self._docstring_diff(),
        ]

    def _write_patch_file(self, lines_to_write: list[str]) -> None:
        r"""Write lines_to_write to a the file called patch_file.

        Parameters
        ----------
        lines_to_write : list[str]
            lines to write to the file - they should be \n terminated
        """
        with open(
            f"{os.path.basename(self.input_file)}.patch", "w", encoding="utf-8"
        ) as file:
            file.writelines(lines_to_write)

    def _overwrite_source_file(self) -> None:
        r"""Overwrite the file with line_to_write.

        Parameters
        ----------
        lines_to_write : list[str]
            lines to write to the file - they should be \n terminated
        """
        tmp_filename = f"{self.input_file}.writing"
        ok = False
        try:
            with open(tmp_filename, "w", encoding="utf-8") as file:
                file.writelines(self._output.lines)
            ok = True
        finally:
            if ok:
                if platform.system() == "Windows":
                    self._windows_rename(tmp_filename)
                else:
                    os.rename(tmp_filename, self.input_file)
            else:
                os.unlink(tmp_filename)

    def _windows_rename(self, tmp_filename: str) -> None:
        """Workaround the fact that os.rename raises an OSError on Windows.

        Parameters
        ----------
        tmp_filename : str
            The file to rename
        """
        if os.path.isfile(self.input_file):
            os.remove(self.input_file)
        os.rename(tmp_filename, self.input_file)

    def report_issues(self) -> tuple[int, str]:
        """Produce a report of all found issues with the docstrings in the file.

        Returns
        -------
        tuple[int, str]
            The number of elements that had issues as well as
            a string representation of those.
        """
        issues: list[str] = []
        for elem in self.docs_list:
            n_issues, report = elem.report_issues()
            if n_issues:
                issues.append(report)
        if not issues:
            return 0, ""
        report = (
            f"The following issues were found in file {self.input_file}:\n"
            + "\n".join(issues)
        )
        return len(issues), report



"""_summary_."""
@overload
def get_expr_as_string(self, annotation: None) -> None:
    ...

@overload
def get_expr_as_string(self, annotation: ast.expr) -> str:
    ...

def get_expr_as_string(self, annotation: Optional[ast.expr]) -> Optional[str]:
    """Turn an expression ast node into a sensible string.

    Used for default values and annotations.
    """
    return None if annotation is None else ast.unparse(annotation)



#!/usr/bin/python

import unittest

def my_func_full(first, second=None, third="value"):
    """This is a description of a method.
    It is on several lines.
    Several styles exists:
      -javadoc,
      -reST,
      -groups.
    It uses the javadoc style.

    @param first: the 1st argument.
    with multiple lines
    @type first: str
    @param second: the 2nd argument.
    @return: the result value
    @rtype: int
    @raise KeyError: raises an exception

    """
    print ("this is the code of my full func")

def my_func_multiline_elem(first,
                           second,
                           third=""):
    '''multiline'''
    print ("this has elems in several lines")

def my_func_empty_params(first, second=None, third="value"):
    print ("this is the code of my empty func")

def my_func_empty_no_params():
    print ("this is the code of my empty func no params")

def my_func_desc_lines_params(first, second=None, third="value"):
    '''This is a description but params not described.
    And a new line!
    '''
    print ("this is the code of my func")

def my_func_desc_line_params(first, second=None, third="value"):
    '''This is a description but params not described.
    '''
    print ("this is the code of my func")

def my_func_desc_line_no_params():
    '''This is a description but no params.
    '''
    print ("this is the code of my func")


def my_func_groups_style(first, second=None, third="value"):
    '''My desc of groups style.

    Parameters:
      first: the first param
      second: the 2nd param
      third: the 3rd param

    '''
    print("group style!")


class MyTestClass(object):

    def __init__(self, one_param=None):
        '''The init with one param

        @param one_param:
        '''
        self.param = one_param
        print("init")

    @classmethod
    def my_cls_method(cls):
        print("cls method")

    def my_method_no_param(self):
        """Do something
        perhaps!
        """
        print("or not")

    def my_method_params_no_docs(self, first, second=None, third="value"):
        print("method params no docs")

    def my_method_multiline_shift_elem(self, first,
                                             second,
                                             third="",
                                             **kwargs):
        '''there are multilines, shift and kwargs'''
        print ("this has elems in several lines")

    def my_method_full(self, first, second=None, third="value"):
        '''The desctiption of method with 3 params and return value
        on several lines

        @param first: the 1st param
        @type first: int
        @param second: the 2nd param with default at None
        @type first: tuple
        @param third: the 3rd param
        @type third: str
        @return: the value
        @rtype: str
        @raise KeyError: key exception

        '''
        print("method full")
        return third
